/* ============================================================

Copyright (c) 2009 Advanced Micro Devices, Inc.  All rights reserved.

Redistribution and use of this material is permitted under the following 
conditions:

Redistributions must retain the above copyright notice and all terms of this 
license.

In no event shall anyone redistributing or accessing or using this material 
commence or participate in any arbitration or legal action relating to this 
material against Advanced Micro Devices, Inc. or any copyright holders or 
contributors. The foregoing shall survive any expiration or termination of 
this license or any agreement or access or use related to this material. 

ANY BREACH OF ANY TERM OF THIS LICENSE SHALL RESULT IN THE IMMEDIATE REVOCATION 
OF ALL RIGHTS TO REDISTRIBUTE, ACCESS OR USE THIS MATERIAL.

THIS MATERIAL IS PROVIDED BY ADVANCED MICRO DEVICES, INC. AND ANY COPYRIGHT 
HOLDERS AND CONTRIBUTORS "AS IS" IN ITS CURRENT CONDITION AND WITHOUT ANY 
REPRESENTATIONS, GUARANTEE, OR WARRANTY OF ANY KIND OR IN ANY WAY RELATED TO 
SUPPORT, INDEMNITY, ERROR FREE OR UNINTERRUPTED OPERA TION, OR THAT IT IS FREE 
FROM DEFECTS OR VIRUSES.  ALL OBLIGATIONS ARE HEREBY DISCLAIMED - WHETHER 
EXPRESS, IMPLIED, OR STATUTORY - INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED 
WARRANTIES OF TITLE, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, 
ACCURACY, COMPLETENESS, OPERABILITY, QUALITY OF SERVICE, OR NON-INFRINGEMENT. 
IN NO EVENT SHALL ADVANCED MICRO DEVICES, INC. OR ANY COPYRIGHT HOLDERS OR 
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, PUNITIVE,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, REVENUE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED OR BASED ON ANY THEORY OF LIABILITY 
ARISING IN ANY WAY RELATED TO THIS MATERIAL, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE. THE ENTIRE AND AGGREGATE LIABILITY OF ADVANCED MICRO DEVICES, 
INC. AND ANY COPYRIGHT HOLDERS AND CONTRIBUTORS SHALL NOT EXCEED TEN DOLLARS 
(US $10.00). ANYONE REDISTRIBUTING OR ACCESSING OR USING THIS MATERIAL ACCEPTS 
THIS ALLOCATION OF RISK AND AGREES TO RELEASE ADVANCED MICRO DEVICES, INC. AND 
ANY COPYRIGHT HOLDERS AND CONTRIBUTORS FROM ANY AND ALL LIABILITIES, 
OBLIGATIONS, CLAIMS, OR DEMANDS IN EXCESS OF TEN DOLLARS (US $10.00). THE 
FOREGOING ARE ESSENTIAL TERMS OF THIS LICENSE AND, IF ANY OF THESE TERMS ARE 
CONSTRUED AS UNENFORCEABLE, FAIL IN ESSENTIAL PURPOSE, OR BECOME VOID OR 
DETRIMENTAL TO ADVANCED MICRO DEVICES, INC. OR ANY COPYRIGHT HOLDERS OR 
CONTRIBUTORS FOR ANY REASON, THEN ALL RIGHTS TO REDISTRIBUTE, ACCESS OR USE 
THIS MATERIAL SHALL TERMINATE IMMEDIATELY. MOREOVER, THE FOREGOING SHALL 
SURVIVE ANY EXPIRATION OR TERMINATION OF THIS LICENSE OR ANY AGREEMENT OR 
ACCESS OR USE RELATED TO THIS MATERIAL.

NOTICE IS HEREBY PROVIDED, AND BY REDISTRIBUTING OR ACCESSING OR USING THIS 
MATERIAL SUCH NOTICE IS ACKNOWLEDGED, THAT THIS MATERIAL MAY BE SUBJECT TO 
RESTRICTIONS UNDER THE LAWS AND REGULATIONS OF THE UNITED STATES OR OTHER 
COUNTRIES, WHICH INCLUDE BUT ARE NOT LIMITED TO, U.S. EXPORT CONTROL LAWS SUCH 
AS THE EXPORT ADMINISTRATION REGULATIONS AND NATIONAL SECURITY CONTROLS AS 
DEFINED THEREUNDER, AS WELL AS STATE DEPARTMENT CONTROLS UNDER THE U.S. 
MUNITIONS LIST. THIS MATERIAL MAY NOT BE USED, RELEASED, TRANSFERRED, IMPORTED,
EXPORTED AND/OR RE-EXPORTED IN ANY MANNER PROHIBITED UNDER ANY APPLICABLE LAWS, 
INCLUDING U.S. EXPORT CONTROL LAWS REGARDING SPECIFICALLY DESIGNATED PERSONS, 
COUNTRIES AND NATIONALS OF COUNTRIES SUBJECT TO NATIONAL SECURITY CONTROLS. 
MOREOVER, THE FOREGOING SHALL SURVIVE ANY EXPIRATION OR TERMINATION OF ANY 
LICENSE OR AGREEMENT OR ACCESS OR USE RELATED TO THIS MATERIAL.

NOTICE REGARDING THE U.S. GOVERNMENT AND DOD AGENCIES: This material is 
provided with "RESTRICTED RIGHTS" and/or "LIMITED RIGHTS" as applicable to 
computer software and technical data, respectively. Use, duplication, 
distribution or disclosure by the U.S. Government and/or DOD agencies is 
subject to the full extent of restrictions in all applicable regulations, 
including those found at FAR52.227 and DFARS252.227 et seq. and any successor 
regulations thereof. Use of this material by the U.S. Government and/or DOD 
agencies is acknowledgment of the proprietary rights of any copyright holders 
and contributors, including those of Advanced Micro Devices, Inc., as well as 
the provisions of FAR52.227-14 through 23 regarding privately developed and/or 
commercial computer software.

This license forms the entire agreement regarding the subject matter hereof and 
supersedes all proposals and prior discussions and writings between the parties 
with respect thereto. This license does not affect any ownership, rights, title,
or interest in, or relating to, this material. No terms of this license can be 
modified or waived, and no breach of this license can be excused, unless done 
so in a writing signed by all affected parties. Each term of this license is 
separately enforceable. If any term of this license is determined to be or 
becomes unenforceable or illegal, such term shall be reformed to the minimum 
extent necessary in order for this license to remain in effect in accordance 
with its terms as modified by such reformation. This license shall be governed 
by and construed in accordance with the laws of the State of Texas without 
regard to rules on conflicts of law of any state or jurisdiction or the United 
Nations Convention on the International Sale of Goods. All disputes arising out 
of this license shall be subject to the jurisdiction of the federal and state 
courts in Austin, Texas, and all defenses are hereby waived concerning personal 
jurisdiction and venue of these courts.

============================================================ */


#include "BoxFilterGLSAT.hpp"
#include "BoxFilterGLSeparable.hpp"
#include <cmath>


int 
main(int argc, char* argv[])
{
    /* By default SAT version runs */
    /* If verification option is given then both versions are verified (SAT and Separable) */
    /* If sep is mentioned with verification option then only Separable version runs */
    /* If sat is mentioned with verification option then only SAT version runs */
    bool flag_sat = false;
    bool flag_sep = false;
    bool flag_verify = false;
    bool separable_verification = false;
    bool sat_verification = false;

    /* Process command line options */
    for(int i = 0; i < argc; i++)
    {
        if(!strcmp(argv[i], "-sep"))
        {
            flag_sep = true;
        }
        if(!strcmp(argv[i], "-sat"))
        {
            flag_sat = true;
        }
        if(!strcmp(argv[i], "-e"))
        {
            flag_verify = true;
        }
    }
    
    /* If -sep is specified with -e then only Separable version's verification runs */
    if(flag_sep == true && flag_verify == true)
        separable_verification = true;

    /* If -sat is specified with -e then only SAT version's verification runs */
    if(flag_sat == true && flag_verify == true)
        sat_verification = true;

    /* If nothing is specified then by default SAT version runs */
    if(flag_sep == false && flag_sat == false)
        flag_sat = true;

    /* Initialize glut only once */
    glutInit(&argc, argv);

    if((flag_sat || flag_verify) && !separable_verification)
    {
        std::cout << "Running SAT version.. " << std::endl;

        BoxFilterGLSAT clBoxFilterGLSAT("OpenCL BoxFilterGLSAT");
        BoxFilterGLSAT::boxFilterGLSAT = &clBoxFilterGLSAT;

        if(clBoxFilterGLSAT.initializeGL(argc, argv) !=  SDK_SUCCESS)
            return SDK_FAILURE;

        if(clBoxFilterGLSAT.initialize() !=  SDK_SUCCESS)
            return SDK_FAILURE;

        if(!clBoxFilterGLSAT.parseCommandLine(argc, argv))
            return SDK_FAILURE;

        if(clBoxFilterGLSAT.isDumpBinaryEnabled())
        {
            return clBoxFilterGLSAT.genBinaryImage();
        }
        else
        {
            cl_int retValue = clBoxFilterGLSAT.setup();
            if(retValue != SDK_SUCCESS)
                return (retValue == SDK_EXPECTED_FAILURE) ? SDK_SUCCESS : SDK_FAILURE;

            if(clBoxFilterGLSAT.run() !=  SDK_SUCCESS)
                return SDK_FAILURE;

            if(clBoxFilterGLSAT.verifyResults() !=  SDK_SUCCESS)
                return SDK_FAILURE;
            if(clBoxFilterGLSAT.cleanup() !=  SDK_SUCCESS)
                return SDK_FAILURE;
        }

        clBoxFilterGLSAT.printStats();
    }
    if(flag_sep || flag_verify && !sat_verification)
    {
        std::cout << "Running Separable version.. " << std::endl;


        BoxFilterGLSeparable clBoxFilterGLSeparable("OpenCL BoxFilterGLSeparable");
        BoxFilterGLSeparable::boxFilterGLSeparable = &clBoxFilterGLSeparable;

        if(clBoxFilterGLSeparable.initializeGL(argc, argv) !=  SDK_SUCCESS)
            return SDK_FAILURE;

        if(clBoxFilterGLSeparable.initialize() !=  SDK_SUCCESS)
            return SDK_FAILURE;

        if(!clBoxFilterGLSeparable.parseCommandLine(argc, argv))
            return SDK_FAILURE;

        if(clBoxFilterGLSeparable.isDumpBinaryEnabled())
        {
            return clBoxFilterGLSeparable.genBinaryImage();
        }
        else
        {

            cl_int retValue = clBoxFilterGLSeparable.setup();
            if(retValue != SDK_SUCCESS)
                return (retValue == SDK_EXPECTED_FAILURE) ? SDK_SUCCESS : SDK_FAILURE;

            if(clBoxFilterGLSeparable.run() !=  SDK_SUCCESS)
                return SDK_FAILURE;

            if(clBoxFilterGLSeparable.verifyResults() !=  SDK_SUCCESS)
                return SDK_FAILURE;
            if(clBoxFilterGLSeparable.cleanup() !=  SDK_SUCCESS)
                return SDK_FAILURE;

            clBoxFilterGLSeparable.printStats();
        }
    }

    return SDK_SUCCESS;
}



