/* ============================================================

Copyright (c) 2011 Advanced Micro Devices, Inc.  All rights reserved.
 
Redistribution and use of this material is permitted under the following 
conditions:
 
Redistributions must retain the above copyright notice and all terms of this 
license.
 
In no event shall anyone redistributing or accessing or using this material 
commence or participate in any arbitration or legal action relating to this 
material against Advanced Micro Devices, Inc. or any copyright holders or 
contributors. The foregoing shall survive any expiration or termination of 
this license or any agreement or access or use related to this material. 

ANY BREACH OF ANY TERM OF THIS LICENSE SHALL RESULT IN THE IMMEDIATE REVOCATION 
OF ALL RIGHTS TO REDISTRIBUTE, ACCESS OR USE THIS MATERIAL.

THIS MATERIAL IS PROVIDED BY ADVANCED MICRO DEVICES, INC. AND ANY COPYRIGHT 
HOLDERS AND CONTRIBUTORS "AS IS" IN ITS CURRENT CONDITION AND WITHOUT ANY 
REPRESENTATIONS, GUARANTEE, OR WARRANTY OF ANY KIND OR IN ANY WAY RELATED TO 
SUPPORT, INDEMNITY, ERROR FREE OR UNINTERRUPTED OPERA TION, OR THAT IT IS FREE 
FROM DEFECTS OR VIRUSES.  ALL OBLIGATIONS ARE HEREBY DISCLAIMED - WHETHER 
EXPRESS, IMPLIED, OR STATUTORY - INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED 
WARRANTIES OF TITLE, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, 
ACCURACY, COMPLETENESS, OPERABILITY, QUALITY OF SERVICE, OR NON-INFRINGEMENT. 
IN NO EVENT SHALL ADVANCED MICRO DEVICES, INC. OR ANY COPYRIGHT HOLDERS OR 
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, PUNITIVE,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, REVENUE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED OR BASED ON ANY THEORY OF LIABILITY 
ARISING IN ANY WAY RELATED TO THIS MATERIAL, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE. THE ENTIRE AND AGGREGATE LIABILITY OF ADVANCED MICRO DEVICES, 
INC. AND ANY COPYRIGHT HOLDERS AND CONTRIBUTORS SHALL NOT EXCEED TEN DOLLARS 
(US $10.00). ANYONE REDISTRIBUTING OR ACCESSING OR USING THIS MATERIAL ACCEPTS 
THIS ALLOCATION OF RISK AND AGREES TO RELEASE ADVANCED MICRO DEVICES, INC. AND 
ANY COPYRIGHT HOLDERS AND CONTRIBUTORS FROM ANY AND ALL LIABILITIES, 
OBLIGATIONS, CLAIMS, OR DEMANDS IN EXCESS OF TEN DOLLARS (US $10.00). THE 
FOREGOING ARE ESSENTIAL TERMS OF THIS LICENSE AND, IF ANY OF THESE TERMS ARE 
CONSTRUED AS UNENFORCEABLE, FAIL IN ESSENTIAL PURPOSE, OR BECOME VOID OR 
DETRIMENTAL TO ADVANCED MICRO DEVICES, INC. OR ANY COPYRIGHT HOLDERS OR 
CONTRIBUTORS FOR ANY REASON, THEN ALL RIGHTS TO REDISTRIBUTE, ACCESS OR USE 
THIS MATERIAL SHALL TERMINATE IMMEDIATELY. MOREOVER, THE FOREGOING SHALL 
SURVIVE ANY EXPIRATION OR TERMINATION OF THIS LICENSE OR ANY AGREEMENT OR 
ACCESS OR USE RELATED TO THIS MATERIAL.

NOTICE IS HEREBY PROVIDED, AND BY REDISTRIBUTING OR ACCESSING OR USING THIS 
MATERIAL SUCH NOTICE IS ACKNOWLEDGED, THAT THIS MATERIAL MAY BE SUBJECT TO 
RESTRICTIONS UNDER THE LAWS AND REGULATIONS OF THE UNITED STATES OR OTHER 
COUNTRIES, WHICH INCLUDE BUT ARE NOT LIMITED TO, U.S. EXPORT CONTROL LAWS SUCH 
AS THE EXPORT ADMINISTRATION REGULATIONS AND NATIONAL SECURITY CONTROLS AS 
DEFINED THEREUNDER, AS WELL AS STATE DEPARTMENT CONTROLS UNDER THE U.S. 
MUNITIONS LIST. THIS MATERIAL MAY NOT BE USED, RELEASED, TRANSFERRED, IMPORTED,
EXPORTED AND/OR RE-EXPORTED IN ANY MANNER PROHIBITED UNDER ANY APPLICABLE LAWS, 
INCLUDING U.S. EXPORT CONTROL LAWS REGARDING SPECIFICALLY DESIGNATED PERSONS, 
COUNTRIES AND NATIONALS OF COUNTRIES SUBJECT TO NATIONAL SECURITY CONTROLS. 
MOREOVER, THE FOREGOING SHALL SURVIVE ANY EXPIRATION OR TERMINATION OF ANY 
LICENSE OR AGREEMENT OR ACCESS OR USE RELATED TO THIS MATERIAL.

NOTICE REGARDING THE U.S. GOVERNMENT AND DOD AGENCIES: This material is 
provided with "RESTRICTED RIGHTS" and/or "LIMITED RIGHTS" as applicable to 
computer software and technical data, respectively. Use, duplication, 
distribution or disclosure by the U.S. Government and/or DOD agencies is 
subject to the full extent of restrictions in all applicable regulations, 
including those found at FAR52.227 and DFARS252.227 et seq. and any successor 
regulations thereof. Use of this material by the U.S. Government and/or DOD 
agencies is acknowledgment of the proprietary rights of any copyright holders 
and contributors, including those of Advanced Micro Devices, Inc., as well as 
the provisions of FAR52.227-14 through 23 regarding privately developed and/or 
commercial computer software.

This license forms the entire agreement regarding the subject matter hereof and 
supersedes all proposals and prior discussions and writings between the parties 
with respect thereto. This license does not affect any ownership, rights, title,
or interest in, or relating to, this material. No terms of this license can be 
modified or waived, and no breach of this license can be excused, unless done 
so in a writing signed by all affected parties. Each term of this license is 
separately enforceable. If any term of this license is determined to be or 
becomes unenforceable or illegal, such term shall be reformed to the minimum 
extent necessary in order for this license to remain in effect in accordance 
with its terms as modified by such reformation. This license shall be governed 
by and construed in accordance with the laws of the State of Texas without 
regard to rules on conflicts of law of any state or jurisdiction or the United 
Nations Convention on the International Sale of Goods. All disputes arising out 
of this license shall be subject to the jurisdiction of the federal and state 
courts in Austin, Texas, and all defenses are hereby waived concerning personal 
jurisdiction and venue of these courts.

============================================================ */


#include "BufferBandwidth.hpp"
#include <string>

struct _memFlags {cl_mem_flags f;
                  const char *s;} 
                  memFlags[] = 
                  {
                     { CL_MEM_READ_ONLY,              "CL_MEM_READ_ONLY" },
                     { CL_MEM_WRITE_ONLY,             "CL_MEM_WRITE_ONLY" },
                     { CL_MEM_READ_WRITE,             "CL_MEM_READ_WRITE" },
                     { CL_MEM_USE_HOST_PTR,           "CL_MEM_USE_HOST_PTR" },
                     { CL_MEM_COPY_HOST_PTR,          "CL_MEM_COPY_HOST_PTR" },
                     { CL_MEM_ALLOC_HOST_PTR,         "CL_MEM_ALLOC_HOST_PTR" },
                     { CL_MEM_USE_PERSISTENT_MEM_AMD, "CL_MEM_USE_PERSISTENT_MEM_AMD"}
                  };

int nFlags = sizeof(memFlags) / sizeof(memFlags[0]);

int 
BufferBandwidth::setupBufferBandwidth()
{

    // Increase the input valuse if -e option is not used
    if(!verify)
    {
        nLoops = 20;
        nRepeats = 1;
        nSkip = 2;
        nKLoops = 20;
        nBytes = 32 * 1024 * 1024;
    }

    tlog = new TestLog(nLoops * nRepeats * 50);
    if(tlog == NULL)
    { 
        sampleCommon->error("Failed to allocate host memory. (tlog)");
        return SDK_FAILURE;
    }

    // educated guess of optimal work size
    int minBytes = MAX_WAVEFRONT_SIZE * sizeof(cl_uint) * 4;
    nBytes = (nBytes / minBytes) * minBytes;
    nBytes = nBytes < minBytes ? minBytes : nBytes;
    nItems = nBytes / (4 * sizeof(cl_uint));
    
    int maxThreads = nBytes / (4 * sizeof( cl_uint));
    nThreads = deviceMaxComputeUnits * numWavefronts * MAX_WAVEFRONT_SIZE;
    if(nThreads > maxThreads)
        nThreads = maxThreads;
    else
    {
        while(nItems % nThreads != 0)
            nThreads += MAX_WAVEFRONT_SIZE;
    }
    nBytesResult = (nThreads / MAX_WAVEFRONT_SIZE) * sizeof(cl_uint) * 2;

#ifdef _WIN32
    memIn = (void *)_aligned_malloc(nBytes, nAlign);
    memOut = (void *)_aligned_malloc(nBytes, nAlign);
    memResult = (void *)_aligned_malloc(nBytesResult, nAlign);
    memRW = (void *)_aligned_malloc(nBytes, nAlign);
#else
    memIn = (void *)memalign(nAlign, nBytes);
    memOut = (void *)memalign(nAlign, nBytes);
    memResult = (void *)memalign(nBytesResult, nBytesResult);
    memRW = (void *)memalign(nAlign, nBytes);
#endif

   if(memIn == NULL ||
      memOut == NULL ||
      memResult == NULL ||
      memRW == NULL) 
   {
        sampleCommon->error("Failed to allocate host memory.(memIn | memOut | memResult | memRW)");
        return SDK_FAILURE;
   }

    return SDK_SUCCESS;
}

int 
BufferBandwidth::genBinaryImage()
{
    cl_int status = CL_SUCCESS;

    /*
     * Have a look at the available platforms and pick either
     * the AMD one if available or a reasonable default.
     */
    cl_uint numPlatforms;
    cl_platform_id platform = NULL;
    status = clGetPlatformIDs(0, NULL, &numPlatforms);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetPlatformIDs failed."))
    {
        return SDK_FAILURE;
    }
    if (0 < numPlatforms) 
    {
        cl_platform_id* platforms = new cl_platform_id[numPlatforms];
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clGetPlatformIDs failed."))
        {
            return SDK_FAILURE;
        }

        char platformName[100];
        for (unsigned i = 0; i < numPlatforms; ++i) 
        {
            status = clGetPlatformInfo(platforms[i],
                                       CL_PLATFORM_VENDOR,
                                       sizeof(platformName),
                                       platformName,
                                       NULL);

            if(!sampleCommon->checkVal(status,
                                       CL_SUCCESS,
                                       "clGetPlatformInfo failed."))
            {
                return SDK_FAILURE;
            }

            platform = platforms[i];
            if (!strcmp(platformName, "Advanced Micro Devices, Inc.")) 
            {
                break;
            }
        }
        std::cout << "Platform found : " << platformName << "\n";
        delete[] platforms;
    }

    if(NULL == platform)
    {
        sampleCommon->error("NULL platform found so Exiting Application.");
        return SDK_FAILURE;
    }

    /*
     * If we could find our platform, use it. Otherwise use just available platform.
     */
    cl_context_properties cps[5] = 
    {
        CL_CONTEXT_PLATFORM, 
        (cl_context_properties)platform, 
        CL_CONTEXT_OFFLINE_DEVICES_AMD,
        (cl_context_properties)1,
        0
    };

    context = clCreateContextFromType(cps,
                                      CL_DEVICE_TYPE_ALL,
                                      NULL,
                                      NULL,
                                      &status);

    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clCreateContextFromType failed."))
    {
        return SDK_FAILURE;
    }

    /* create a CL program using the kernel source */
    streamsdk::SDKFile kernelFile;
    std::string kernelPath = sampleCommon->getPath();
    kernelPath.append("BufferBandwidth_Kernels.cl");
    if(!kernelFile.open(kernelPath.c_str()))
    {
        std::cout << "Failed to load kernel file : " << kernelPath << std::endl;
        return SDK_FAILURE;
    }
    const char * source = kernelFile.source().c_str();
    size_t sourceSize[] = {strlen(source)};
    program = clCreateProgramWithSource(context,
                                        1,
                                        &source,
                                        sourceSize,
                                        &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clCreateProgramWithSource failed."))
    {
        return SDK_FAILURE;
    }
    
    
    std::string flagsStr = std::string("");

    // Get additional options
    if(isComplierFlagsSpecified())
    {
        streamsdk::SDKFile flagsFile;
        std::string flagsPath = sampleCommon->getPath();
        flagsPath.append(flags.c_str());
        if(!flagsFile.open(flagsPath.c_str()))
        {
            std::cout << "Failed to load flags file: " << flagsPath << std::endl;
            return SDK_FAILURE;
        }
        flagsFile.replaceNewlineWithSpaces();
        const char * flags = flagsFile.source().c_str();
        flagsStr.append(flags);
    }

    if(flagsStr.size() != 0)
        std::cout << "Build Options are : " << flagsStr.c_str() << std::endl;


    /* create a cl program executable for all the devices specified */
    status = clBuildProgram(program,
                            0,
                            NULL,
                            flagsStr.c_str(),
                            NULL,
                            NULL);

    sampleCommon->checkVal(status,
                        CL_SUCCESS,
                        "clBuildProgram failed.");

    size_t numDevices;
    status = clGetProgramInfo(program, 
                           CL_PROGRAM_NUM_DEVICES,
                           sizeof(numDevices),
                           &numDevices,
                           NULL );
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_NUM_DEVICES) failed."))
    {
        return SDK_FAILURE;
    }

    std::cout << "Number of devices found : " << numDevices << "\n\n";
    devices = (cl_device_id *)malloc( sizeof(cl_device_id) * numDevices );
    if(devices == NULL)
    {
        sampleCommon->error("Failed to allocate host memory.(devices)");
        return SDK_FAILURE;
    }
    /* grab the handles to all of the devices in the program. */
    status = clGetProgramInfo(program, 
                              CL_PROGRAM_DEVICES, 
                              sizeof(cl_device_id) * numDevices,
                              devices,
                              NULL );
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_DEVICES) failed."))
    {
        return SDK_FAILURE;
    }


    /* figure out the sizes of each of the binaries. */
    size_t *binarySizes = (size_t*)malloc( sizeof(size_t) * numDevices );
    if(devices == NULL)
    {
        sampleCommon->error("Failed to allocate host memory.(binarySizes)");
        return SDK_FAILURE;
    }
    
    status = clGetProgramInfo(program, 
                              CL_PROGRAM_BINARY_SIZES,
                              sizeof(size_t) * numDevices, 
                              binarySizes, NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_BINARY_SIZES) failed."))
    {
        return SDK_FAILURE;
    }

    size_t i = 0;
    /* copy over all of the generated binaries. */
    char **binaries = (char **)malloc( sizeof(char *) * numDevices );
    if(binaries == NULL)
    {
        sampleCommon->error("Failed to allocate host memory.(binaries)");
        return SDK_FAILURE;
    }

    for(i = 0; i < numDevices; i++)
    {
        if(binarySizes[i] != 0)
        {
            binaries[i] = (char *)malloc( sizeof(char) * binarySizes[i]);
            if(binaries[i] == NULL)
            {
                sampleCommon->error("Failed to allocate host memory.(binaries[i])");
                return SDK_FAILURE;
            }
        }
        else
        {
            binaries[i] = NULL;
        }
    }
    status = clGetProgramInfo(program, 
                              CL_PROGRAM_BINARIES,
                              sizeof(char *) * numDevices, 
                              binaries, 
                              NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_BINARIES) failed."))
    {
        return SDK_FAILURE;
    }

    /* dump out each binary into its own separate file. */
    for(i = 0; i < numDevices; i++)
    {
        char fileName[100];
        sprintf(fileName, "%s.%d", dumpBinary.c_str(), (int)i);
        if(binarySizes[i] != 0)
        {
            char deviceName[1024];
            status = clGetDeviceInfo(devices[i], 
                                     CL_DEVICE_NAME, 
                                     sizeof(deviceName),
                                     deviceName, 
                                     NULL);
            if(!sampleCommon->checkVal(status,
                                       CL_SUCCESS,
                                       "clGetDeviceInfo(CL_DEVICE_NAME) failed."))
            {
                return SDK_FAILURE;
            }

            printf( "%s binary kernel: %s\n", deviceName, fileName);
            streamsdk::SDKFile BinaryFile;
            if(!BinaryFile.writeBinaryToFile(fileName, 
                                             binaries[i], 
                                             binarySizes[i]))
            {
                std::cout << "Failed to load kernel file : " << fileName << std::endl;
                return SDK_FAILURE;
            }
        }
        else
        {
            printf("Skipping %s since there is no binary data to write!\n",
                    fileName);
        }
    }

    // Release all resouces and memory
    for(i = 0; i < numDevices; i++)
    {
        if(binaries[i] != NULL)
        {
            free(binaries[i]);
            binaries[i] = NULL;
        }
    }

    if(binaries != NULL)
    {
        free(binaries);
        binaries = NULL;
    }

    if(binarySizes != NULL)
    {
        free(binarySizes);
        binarySizes = NULL;
    }

    if(devices != NULL)
    {
        free(devices);
        devices = NULL;
    }

    status = clReleaseProgram(program);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clReleaseProgram failed."))
    {
        return SDK_FAILURE;
    }

    status = clReleaseContext(context);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clReleaseContext failed."))
    {
        return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}


int
BufferBandwidth::setupCL(void)
{
    cl_int status = 0;
    size_t deviceListSize;

    cl_device_type dType;
    
    if(deviceType.compare("cpu") == 0)
    {
        dType = CL_DEVICE_TYPE_CPU;
    }
    else //deviceType = "gpu" 
    {
        dType = CL_DEVICE_TYPE_GPU;
        if(isThereGPU() == false)
        {
            std::cout << "GPU not found. Falling back to CPU device" << std::endl;
            dType = CL_DEVICE_TYPE_CPU;
        }
    }

    /*
     * Have a look at the available platforms and pick either
     * the AMD one if available or a reasonable default.
     */

    cl_uint numPlatforms;
    cl_platform_id platform = NULL;
    status = clGetPlatformIDs(0, NULL, &numPlatforms);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetPlatformIDs failed."))
    {
        return SDK_FAILURE;
    }
    if(0 < numPlatforms) 
    {
        cl_platform_id* platforms = new cl_platform_id[numPlatforms];
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clGetPlatformIDs failed."))
        {
            return SDK_FAILURE;
        }

        if(isPlatformEnabled())
        {
            platform = platforms[platformId];
        }
        else
        {
            for(unsigned i = 0; i < numPlatforms; ++i) 
            {
                char pbuf[100];
                status = clGetPlatformInfo(platforms[i],
                                           CL_PLATFORM_VENDOR,
                                           sizeof(pbuf),
                                           pbuf,
                                           NULL);

                if(!sampleCommon->checkVal(status,
                                           CL_SUCCESS,
                                           "clGetPlatformInfo failed."))
                {
                    return SDK_FAILURE;
                }

                platform = platforms[i];
                if (!strcmp(pbuf, "Advanced Micro Devices, Inc.")) 
                {
                    break;
                }
            }
        }
        delete[] platforms;
    }

    if(NULL == platform)
    {
        sampleCommon->error("NULL platform found so Exiting Application.");
        return SDK_FAILURE;
    }

    // Display available devices.
    if(!sampleCommon->displayDevices(platform, dType))
    {
        sampleCommon->error("sampleCommon::displayDevices() failed");
        return SDK_FAILURE;
    }

    /*
     * If we could find our platform, use it. Otherwise use just available platform.
     */

    cl_context_properties cps[3] = 
    {
        CL_CONTEXT_PLATFORM, 
        (cl_context_properties)platform, 
        0
    };

    context = clCreateContextFromType(cps,
                                      dType,
                                      NULL,
                                      NULL,
                                      &status);
    if(!sampleCommon->checkVal(status, 
                               CL_SUCCESS,
                               "clCreateContextFromType failed."))
        return SDK_FAILURE;

    /* First, get the size of device list data */
    status = clGetContextInfo(context, 
                              CL_CONTEXT_DEVICES, 
                              0, 
                              NULL, 
                              &deviceListSize);
    if(!sampleCommon->checkVal(status, 
                               CL_SUCCESS,
                               "clGetContextInfo failed."))
        return SDK_FAILURE;

    int deviceCount = (int)(deviceListSize / sizeof(cl_device_id));
    if(!sampleCommon->validateDeviceId(deviceId, deviceCount))
    {
        sampleCommon->error("sampleCommon::validateDeviceId() failed");
        return SDK_FAILURE;
    }

    /* Now allocate memory for device list based on the size we got earlier */
    devices = (cl_device_id*)malloc(deviceListSize);
    if(devices == NULL) 
    {
        sampleCommon->error("Failed to allocate memory (devices).");
        return SDK_FAILURE;
    }

    /* Now, get the device list data */
    status = clGetContextInfo(context, 
                              CL_CONTEXT_DEVICES, 
                              deviceListSize, 
                              devices, 
                              NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clGetGetContextInfo failed."))
        return SDK_FAILURE;

    status = clGetDeviceInfo(devices[deviceId],
                           CL_DEVICE_MAX_COMPUTE_UNITS,
                           sizeof(cl_uint),
                           &deviceMaxComputeUnits,
                           NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clGetDeviceInfo CL_DEVICE_MAX_COMPUTE_UNITS failed."))
        return SDK_FAILURE;


    // Check device extensions
    char* deviceExtensions = NULL;;
    size_t extStringSize = 0;

    // Get device extensions 
    status = clGetDeviceInfo(devices[deviceId], 
                            CL_DEVICE_EXTENSIONS,
                            0, 
                            NULL, 
                            &extStringSize);
    if(!sampleCommon->checkVal(status,
                                CL_SUCCESS, 
                                "clGetDeviceInfo failed.(CL_DEVICE_EXTENSIONS)"))
        return SDK_FAILURE;
    
    deviceExtensions = new char[extStringSize];
    if(NULL == deviceExtensions)
    {
        sampleCommon->error("Failed to allocate memory(deviceExtensions)");
        return  SDK_FAILURE;
    }

    status = clGetDeviceInfo(devices[deviceId],
                            CL_DEVICE_EXTENSIONS, 
                            extStringSize,
                            deviceExtensions,
                            NULL);
    if(!sampleCommon->checkVal(status,
                                CL_SUCCESS, 
                                "clGetDeviceInfo failed.(CL_DEVICE_EXTENSIONS)"))
        return SDK_FAILURE;


    /* Check if cl_khr_fp64 extension is supported */
    if(!strstr(deviceExtensions, "cl_khr_local_int32_base_atomics"))
    {
        sampleCommon->expectedError("Device does not support cl_khr_local_int32_base_atomics extension!");
        return SDK_EXPECTED_FAILURE;
    }
    delete deviceExtensions;

    // Get OpenCL device version
    char deviceVersion[1024];
    status = clGetDeviceInfo(devices[deviceId], 
        CL_DEVICE_VERSION, 
        sizeof(deviceVersion), 
        deviceVersion, 
        NULL);

    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS, 
        "clGetDeviceInfo failed.(CL_DEVICE_VERSION)"))
        return SDK_FAILURE;

    std::string deviceVersionStr = std::string(deviceVersion);
    size_t vStart = deviceVersionStr.find(" ", 0);
    size_t vEnd = deviceVersionStr.find(" ", vStart + 1);
    std::string vStrVal = deviceVersionStr.substr(vStart + 1, vEnd - vStart - 1);

    // Check of OPENCL_C_VERSION if device version is 1.1 or later
    if(vStrVal.compare("1.0") > 0)
    {
        //Get OPENCL_C_VERSION 
        status = clGetDeviceInfo(devices[deviceId], 
            CL_DEVICE_OPENCL_C_VERSION, 
            sizeof(deviceVersion), 
            deviceVersion, 
            NULL);

        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clGetDeviceInfo failed.(CL_DEVICE_OPENCL_C_VERSION)"))
            return SDK_FAILURE;

        // Exit if OpenCL C device version is 1.0
        deviceVersionStr = std::string(deviceVersion);
        vStart = deviceVersionStr.find(" ", 0);
        vEnd = deviceVersionStr.find(" ", vStart + 1);
        vStrVal = deviceVersionStr.substr(vStart + 1, vEnd - vStart - 1);
        if(vStrVal.compare("1.0") <= 0)
        {
            sampleCommon->expectedError("Unsupported device! Required CL_DEVICE_OPENCL_C_VERSION as 1.1");
            return SDK_EXPECTED_FAILURE;
        }
    }
    else
    {
        sampleCommon->expectedError("Unsupported device! Required CL_DEVICE_OPENCL_C_VERSION as 1.1");
        return SDK_EXPECTED_FAILURE;
    }

    if(setupBufferBandwidth() != SDK_SUCCESS)
        return SDK_FAILURE;

    {
        /* The block is to move the declaration of prop closer to its use */
        cl_command_queue_properties prop = 0;
#ifdef USE_CL_PROFILER
        prop |= CL_QUEUE_PROFILING_ENABLE;
#endif

        queue = clCreateCommandQueue(context, 
                                            devices[deviceId], 
                                            prop, 
                                            &status);
        if(!sampleCommon->checkVal(status,
                                   0,
                                   "clCreateCommandQueue failed."))
            return SDK_FAILURE;
    }

    void *hostPtr = NULL;

   if(inFlags & CL_MEM_USE_HOST_PTR || inFlags & CL_MEM_COPY_HOST_PTR)
       hostPtr = memIn;

    inputBuffer = clCreateBuffer(context, inFlags, nBytes, hostPtr, &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(inputBuffer)"))
        return SDK_FAILURE;

    hostPtr = NULL;

    if(outFlags & CL_MEM_USE_HOST_PTR || outFlags & CL_MEM_COPY_HOST_PTR)
        hostPtr = memOut;

    outputBuffer = clCreateBuffer(context, outFlags, nBytes, hostPtr, &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(outputBuffer)"))
        return SDK_FAILURE;

    hostPtr = NULL;

    if(copyFlags & CL_MEM_USE_HOST_PTR || copyFlags & CL_MEM_COPY_HOST_PTR)
        hostPtr = memIn;

    if(whichTest == 2 || verify)
    {
        inputCopyBuffer = clCreateBuffer(context, copyFlags, nBytes, hostPtr, &status);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS, 
                                   "clCreateBuffer() failed.(inputCopyBuffer)"))
            return SDK_FAILURE;

        outputCopyBuffer = clCreateBuffer(context, copyFlags, nBytes, hostPtr, &status);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS, 
                                   "clCreateBuffer() failed.(outputCopyBuffer)"))
            return SDK_FAILURE;
    }

    resultBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, nBytesResult, NULL, &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(resultBuffer)"))
        return SDK_FAILURE;


    /* create a CL program using the kernel source */
    streamsdk::SDKFile kernelFile;
    std::string kernelPath = sampleCommon->getPath();

    if(isLoadBinaryEnabled())
    {
        kernelPath.append(loadBinary.c_str());
        if(!kernelFile.readBinaryFromFile(kernelPath.c_str()))
        {
            std::cout << "Failed to load kernel file : " << kernelPath << std::endl;
            return SDK_FAILURE;
        }

        const char * binary = kernelFile.source().c_str();
        size_t binarySize = kernelFile.source().size();
        program = clCreateProgramWithBinary(context,
                                            1,
                                            &devices[deviceId], 
                                            (const size_t *)&binarySize,
                                            (const unsigned char**)&binary,
                                            NULL,
                                            &status);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clCreateProgramWithBinary failed."))
        {
            return SDK_FAILURE;
        }

    }
    else
    {
        kernelPath.append("BufferBandwidth_Kernels.cl");
        if(!kernelFile.open(kernelPath.c_str()))
        {
            std::cout << "Failed to load kernel file: " << kernelPath << std::endl;
            return SDK_FAILURE;
        }
        const char * source = kernelFile.source().c_str();
        size_t sourceSize[] = {strlen(source)};
        program = clCreateProgramWithSource(context,
                                            1,
                                            &source,
                                            sourceSize,
                                            &status);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clCreateProgramWithSource failed."))
            return SDK_FAILURE;
    }

    std::string flagsStr = std::string("");

    // Get additional options
    if(isComplierFlagsSpecified())
    {
        streamsdk::SDKFile flagsFile;
        std::string flagsPath = sampleCommon->getPath();
        flagsPath.append(flags.c_str());
        if(!flagsFile.open(flagsPath.c_str()))
        {
            std::cout << "Failed to load flags file: " << flagsPath << std::endl;
            return SDK_FAILURE;
        }
        flagsFile.replaceNewlineWithSpaces();
        const char * flags = flagsFile.source().c_str();
        flagsStr.append(flags);
    }

    if(flagsStr.size() != 0)
        std::cout << "Build Options are : " << flagsStr.c_str() << std::endl;

    /* create a cl program executable for all the devices specified */
    status = clBuildProgram(program, 1, &devices[deviceId], flagsStr.c_str(), NULL, NULL);
    if(status != CL_SUCCESS)
    {
        if(status == CL_BUILD_PROGRAM_FAILURE)
        {
            cl_int logStatus;
            char *buildLog = NULL;
            size_t buildLogSize = 0;
            logStatus = clGetProgramBuildInfo (program, 
                devices[deviceId], 
                CL_PROGRAM_BUILD_LOG, 
                buildLogSize, 
                buildLog, 
                &buildLogSize);
            if(!sampleCommon->checkVal(
                logStatus,
                CL_SUCCESS,
                "clGetProgramBuildInfo failed."))
                return SDK_FAILURE;

            buildLog = (char*)malloc(buildLogSize);
            if(buildLog == NULL)
            {
                sampleCommon->error("Failed to allocate host memory. (buildLog)");
                return SDK_FAILURE;
            }
            memset(buildLog, 0, buildLogSize);

            logStatus = clGetProgramBuildInfo (program, 
                devices[deviceId], 
                CL_PROGRAM_BUILD_LOG, 
                buildLogSize, 
                buildLog, 
                NULL);
            if(!sampleCommon->checkVal(
                logStatus,
                CL_SUCCESS,
                "clGetProgramBuildInfo failed."))
            {
                free(buildLog);
                return SDK_FAILURE;
            }

            std::cout << " \n\t\t\tBUILD LOG\n";
            std::cout << " ************************************************\n";
            std::cout << buildLog << std::endl;
            std::cout << " ************************************************\n";
            free(buildLog);
        }

        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clBuildProgram failed."))
            return SDK_FAILURE;
    }

/* ConstantBuffer bandwidth from single access */
    readKernel = clCreateKernel(program, "readKernel", &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clCreateKernel failed.(readKernel)"))
        return SDK_FAILURE;

    size_t kernelWorkGroupSize = 0;

    /* Check whether specified local group size is possible on current kernel */
    status = clGetKernelWorkGroupInfo(readKernel,
                                      devices[deviceId],
                                      CL_KERNEL_WORK_GROUP_SIZE,
                                      sizeof(size_t),
                                      &kernelWorkGroupSize,
                                      0);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetKernelWorkGroupInfo failed."))
        return SDK_FAILURE;


    // If local groupSize exceeds the maximum supported on kernel  fall back
    if(MAX_WAVEFRONT_SIZE > kernelWorkGroupSize)
    {
        if(!quiet)
        {
            std::cout << "Out of Resources!" << std::endl;
            std::cout << "Group Size specified : " << MAX_WAVEFRONT_SIZE << std::endl;
            std::cout << "Max Group Size supported on the kernel(readKernel) : " 
                      << kernelWorkGroupSize << std::endl;
            std::cout << "Falling back to " << kernelWorkGroupSize << std::endl;
        }
        return SDK_FAILURE;
    }

    writeKernel = clCreateKernel(program, "writeKernel", &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clCreateKernel failed.(writeKernel)"))
        return SDK_FAILURE;

    // Check whether specified local group size is possible on current kernel
    status = clGetKernelWorkGroupInfo(writeKernel,
                                      devices[deviceId],
                                      CL_KERNEL_WORK_GROUP_SIZE,
                                      sizeof(size_t),
                                      &kernelWorkGroupSize,
                                      0);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clGetKernelWorkGroupInfo failed."))
        return SDK_FAILURE;


    // If local group size exceeds the maximum supported on kernel fall back 
    if(MAX_WAVEFRONT_SIZE > kernelWorkGroupSize)
    {
        if(!quiet)
        {
            std::cout << "Out of Resources!" << std::endl;
            std::cout << "Group Size specified : " << MAX_WAVEFRONT_SIZE << std::endl;
            std::cout << "Max Group Size supported on the kernel(writeKernel) : " 
                      << kernelWorkGroupSize << std::endl;
            std::cout << "Falling back to " << kernelWorkGroupSize << std::endl;
        }
        return SDK_FAILURE;
    }


    return SDK_SUCCESS;
}

int 
BufferBandwidth::initialize()
{
    /* Call base class Initialize to get default configuration */
    if(!this->SDKSample::initialize())
        return SDK_FAILURE;

    streamsdk::Option* numLoops = new streamsdk::Option;

    numLoops->_sVersion = "i";
    numLoops->_lVersion = "iterations";
    numLoops->_description = "Number of timing loops";
    numLoops->_type = streamsdk::CA_ARG_INT;
    numLoops->_value = &nLoops;

    sampleArgs->AddOption(numLoops);
    delete numLoops;

    streamsdk::Option* numRepeats = new streamsdk::Option;

    numRepeats->_sVersion = "r";
    numRepeats->_lVersion = "repeats";
    numRepeats->_description = "Repeat each timing <n> times";
    numRepeats->_type = streamsdk::CA_ARG_INT;
    numRepeats->_value = &nRepeats;

    sampleArgs->AddOption(numRepeats);
    delete numRepeats;

    streamsdk::Option* kernelLoops = new streamsdk::Option;

    kernelLoops->_sVersion = "k";
    kernelLoops->_lVersion = "kernelLoops";
    kernelLoops->_description = "Number of loops in kernel";
    kernelLoops->_type = streamsdk::CA_ARG_INT;
    kernelLoops->_value = &nKLoops;

    sampleArgs->AddOption(kernelLoops);
    delete kernelLoops;

    streamsdk::Option* size = new streamsdk::Option;

    size->_sVersion = "x";
    size->_lVersion = "size";
    size->_description = "Size in bytes";
    size->_type = streamsdk::CA_ARG_INT;
    size->_value = &nBytes;

    sampleArgs->AddOption(size);
    delete size;

    streamsdk::Option* skipLoops = new streamsdk::Option;

    skipLoops->_sVersion = "s";
    skipLoops->_lVersion = "skip";
    skipLoops->_description = "skip first n iterations for average";
    skipLoops->_type = streamsdk::CA_ARG_INT;
    skipLoops->_value = &nSkip;

    sampleArgs->AddOption(skipLoops);
    delete skipLoops;


    streamsdk::Option* wavefronts = new streamsdk::Option;

    wavefronts->_sVersion = "w";
    wavefronts->_lVersion = "wavefronts";
    wavefronts->_description = "Number of wavefronts per compute unit";
    wavefronts->_type = streamsdk::CA_ARG_INT;
    wavefronts->_value = &numWavefronts;

    sampleArgs->AddOption(wavefronts);
    delete wavefronts;

    streamsdk::Option* typeOfTest = new streamsdk::Option;

    typeOfTest->_sVersion = "T";
    typeOfTest->_lVersion = "typeOfTest";
    typeOfTest->_description = "Type of test"
                                "\n\t\t\t 0 for clEnqueue[Map,Unmap]"
                                "\n\t\t\t 1 for clEnqueue[Read/Write]"
                                "\n\t\t\t 2 for clEnqueueCopy\n";
    typeOfTest->_type = streamsdk::CA_ARG_INT;
    typeOfTest->_value = &whichTest;

    sampleArgs->AddOption(typeOfTest);
    delete typeOfTest;

    streamsdk::Option* inputFlags = new streamsdk::Option;

    inputFlags->_sVersion = "I";
    inputFlags->_lVersion = "inMemFlag";
    inputFlags->_description = "Memory flags for input buffer "
                         "\n\t\t 0 for CL_MEM_READ_ONLY"
                         "\n\t\t 1 CL_MEM_WRITE_ONLY"
                         "\n\t\t 2 CL_MEM_READ_WRITE"
                         "\n\t\t 3 CL_MEM_USE_HOST_PTR"
                         "\n\t\t 4 CL_MEM_COPY_HOST_PTR"
                         "\n\t\t 5 CL_MEM_ALLOC_HOST_PTR"
                         "\n\t\t 6 CL_MEM_USE_PERSISTENT_MEM_AMD\n";
    inputFlags->_type = streamsdk::CA_ARG_INT;
    inputFlags->_value = &inFlagsValue;

    sampleArgs->AddOption(inputFlags);
    delete inputFlags;

    streamsdk::Option* outputFlags = new streamsdk::Option;

    outputFlags->_sVersion = "O";
    outputFlags->_lVersion = "outMemFlag";
    outputFlags->_description = "Memory flags for output buffer."
                                "\n\t\t\t Values are same as Option -I\n";
    outputFlags->_type = streamsdk::CA_ARG_INT;
    outputFlags->_value = &outFlagsValue;

    sampleArgs->AddOption(outputFlags);
    delete outputFlags;

    streamsdk::Option* copyFlags = new streamsdk::Option;

    copyFlags->_sVersion = "C";
    copyFlags->_lVersion = "copyMemFlag";
    copyFlags->_description = "Memory flags for copy buffer."
                                "\n\t\t\t Values are same as Option -I\n";
    copyFlags->_type = streamsdk::CA_ARG_INT;
    copyFlags->_value = &copyFlagsValue;

    sampleArgs->AddOption(copyFlags);
    delete copyFlags;

    streamsdk::Option* mapping = new streamsdk::Option;

    mapping->_sVersion = "m";
    mapping->_lVersion = "mapping";
    mapping->_description = "Always maps as MAP_READ | MAP_WRITE";
    mapping->_type = streamsdk::CA_NO_ARGUMENT;
    mapping->_value = &mapRW;

    sampleArgs->AddOption(mapping);
    delete mapping;


    streamsdk::Option* disable = new streamsdk::Option;

    disable->_sVersion = "D";
    disable->_lVersion = "disable";
    disable->_description = "Disable host mem B/W baseline";
    disable->_type = streamsdk::CA_NO_ARGUMENT;
    disable->_value = &doHost;

    sampleArgs->AddOption(disable);
    delete disable;

    streamsdk::Option* pringLogStr = new streamsdk::Option;

    pringLogStr->_sVersion = "l";
    pringLogStr->_lVersion = "log";
    pringLogStr->_description = "Prints complete timing log";
    pringLogStr->_type = streamsdk::CA_NO_ARGUMENT;
    pringLogStr->_value = &printLog;

    sampleArgs->AddOption(pringLogStr);
    delete pringLogStr;

    return SDK_SUCCESS;
}

int 
BufferBandwidth::parseExtraCommandLineOptions(int argc, char**argv)
{
    if(!parseCommandLine(argc, argv))
        return SDK_FAILURE;
    
    //Handle extra options which are not default in SDKUtils
    while(--argc) 
    {
        if(strcmp(argv[argc], "-I") == 0 ||  strcmp(argv[argc], "-inMemFlag") == 0)
        {
            int f = atoi( argv[ argc+1 ] );
            if( f < nFlags )
                inFlags |= memFlags[ f ].f;
        }

        if(strcmp(argv[argc], "-O") == 0 || strcmp(argv[argc], "-outMemFlag") == 0)
        {
            int f = atoi( argv[ argc+1 ] );
            if( f < nFlags )
                outFlags |= memFlags[ f ].f;
        }

        if(strcmp(argv[argc], "-C") == 0 || strcmp(argv[argc], "-copyMemFlag") == 0)
        {
            int f = atoi( argv[ argc+1 ] );
            if( f < nFlags )
                copyFlags |= memFlags[ f ].f;
        }
    }

    cl_mem_flags f = CL_MEM_READ_ONLY |
                     CL_MEM_WRITE_ONLY |
                     CL_MEM_READ_WRITE;

    if((inFlags & f) == 0)
        inFlags |= CL_MEM_READ_ONLY;

    if((outFlags & f) == 0)
        outFlags |= CL_MEM_WRITE_ONLY;

    f |= CL_MEM_USE_HOST_PTR |
         CL_MEM_COPY_HOST_PTR |
         CL_MEM_ALLOC_HOST_PTR |
         CL_MEM_USE_PERSISTENT_MEM_AMD;

    if((copyFlags & f) == 0)
        copyFlags = CL_MEM_ALLOC_HOST_PTR | CL_MEM_READ_WRITE;

    f = CL_MEM_READ_ONLY |
        CL_MEM_WRITE_ONLY |
        CL_MEM_READ_WRITE;

    if((copyFlags & f) == 0)
        copyFlags |= CL_MEM_READ_WRITE;

    nSkip = nLoops > nSkip ? nSkip : 0;

    if(whichTest > 2)
    {
        sampleCommon->error("Only 0 to 2 are valid values for -T option");
        usage();
        return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}

int 
BufferBandwidth::setup()
{
    int status = setupCL();
    if(status != SDK_SUCCESS)
        return status;

    return SDK_SUCCESS;
}

int BufferBandwidth::timedBufMappedRead(cl_mem buf, unsigned char v)
{
    CPerfCounter t1, t2, t3;
    cl_int status;
    cl_event ev;
    void *ptr = NULL;
    cl_map_flags mapFlag = CL_MAP_READ | CL_MAP_WRITE;

    t1.Reset();
    t2.Reset();
    t3.Reset();

    t1.Start();

    if(!mapRW)
        mapFlag = CL_MAP_READ;

    ptr = (void *)clEnqueueMapBuffer(queue,
                                     buf,
                                     CL_TRUE,
                                     mapFlag,
                                     0,
                                     nBytes,
                                     0, 
                                     NULL, 
                                     NULL,
                                     &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueMapBuffer failed.(buf)"))
        return SDK_FAILURE;

    t1.Stop();
    t2.Start();

    bool results = readVerifyMemCPU(ptr, v, nBytes);

    t2.Stop();
    t3.Start();

    status = clEnqueueUnmapMemObject(queue,
                                  buf,
                                  (void *)ptr,
                                  0,
                                  NULL,
                                  &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueUnmapMemObject failed.(buf)"))
        return SDK_FAILURE;

    status = clWaitForEvents(1, &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clWaitForEvents() failed.(ev)"))
        return SDK_FAILURE;

    t3.Stop();

    // Release event
    status = clReleaseEvent(ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseEvent() failed.(ev)"))
        return SDK_FAILURE;

    const char *msg = NULL;
    if(mapRW)
        msg = "clEnqueueMapBuffer(MAP_READ|MAP_WRITE)";
    else
        msg = "clEnqueueMapBuffer(MAP_READ)";

    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", 
        msg, 
        t1.GetElapsedTime(), 
        nBytes, 
        1);
    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "CPU read",
        t2.GetElapsedTime(),
        nBytes,
        1); 

    if(results)
        tlog->Msg("%32s\n", "verification ok");
    else
    {
        correctness = false;
        tlog->Error("%32s\n", "verification FAILED");
        printf("Failed!\n");
        return SDK_FAILURE;
    }

    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clEnqueueUnmapMemObject()", 
        t3.GetElapsedTime(), 
        nBytes,
        1);

    return SDK_SUCCESS;
}

int BufferBandwidth::timedBufMappedWrite(cl_mem buf, unsigned char v)
{
    CPerfCounter t1, t2, t3;
    cl_int status;
    cl_event ev;
    void *ptr = NULL;
    cl_map_flags mapFlag = CL_MAP_READ | CL_MAP_WRITE;

    t1.Reset();
    t2.Reset();
    t3.Reset();

    t1.Start();

    if(!mapRW)
        mapFlag = CL_MAP_WRITE;

    ptr = (void *)clEnqueueMapBuffer(queue,
                                     buf,
                                     CL_TRUE,
                                     mapFlag,
                                     0,
                                     nBytes,
                                     0,
                                     NULL,
                                     NULL,
                                     &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueMapBuffer() failed.(buf)"))
        return SDK_FAILURE;

    t1.Stop();
    t2.Start();

    memset(ptr, v, nBytes);

    t2.Stop();
    t3.Start();

    status = clEnqueueUnmapMemObject(queue,
                                  buf,
                                  (void *)ptr,
                                  0,
                                  NULL,
                                  &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueUnmapMemObject() failed.(buf)"))
        return SDK_FAILURE;

    status = clWaitForEvents(1, &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clWaitForEvents() failed.(ev)"))
        return SDK_FAILURE;

    t3.Stop();

	// Release event
    status = clReleaseEvent(ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseEvent() failed.(ev)"))
        return SDK_FAILURE;

    const char *msg = NULL;

    if(mapRW)
        msg = "clEnqueueMapBuffer(MAP_READ|MAP_WRITE)";
    else
        msg = "clEnqueueMapBuffer(MAP_WRITE)";

    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n",
        msg, 
        t1.GetElapsedTime(), 
        nBytes, 
        1);

    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "memset()", 
        t2.GetElapsedTime(), 
        nBytes, 
        1);    
    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clEnqueueUnmapMemObject()", 
        t3.GetElapsedTime(), 
        nBytes, 
        1);

    return SDK_SUCCESS;
}

int BufferBandwidth::timedBufCLRead(cl_mem buf,
                     void *ptr,
                     unsigned char v)
{
    CPerfCounter t1;
    cl_int status = CL_SUCCESS;

    t1.Start();

    status = clEnqueueReadBuffer(queue,
                              buf,
                              CL_TRUE,
                              0,
                              nBytes,
                              ptr,
                              0,
                              NULL,
                              NULL);

    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueReadBuffer() failed.(buf)"))
        return SDK_FAILURE;

    t1.Stop();

    bool results = readVerifyMemCPU(ptr, v, nBytes);

    tlog->Timer("%32s  %lf s %8.2lf GB/s\n", 
        "clEnqueueReadBuffer", 
        t1.GetElapsedTime(), 
        nBytes,
        1);

    if(results)
        tlog->Msg("%32s\n", "verification ok");
    else
    {
        printf("Failed!\n");
        correctness = false;
        tlog->Error("%32s\n", "verification FAILED");
        return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}

int BufferBandwidth::timedBufCLWrite(cl_mem buf, void *ptr)
{
    CPerfCounter t1;
    cl_int status = CL_SUCCESS;

    t1.Start();

    status = clEnqueueWriteBuffer(queue,
                               buf,
                               CL_TRUE,
                               0,
                               nBytes,
                               ptr,
                               0, 
                               NULL, 
                               NULL);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueWriteBuffer() failed.(buf)"))
        return SDK_FAILURE;

    t1.Stop();
    
    tlog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", 
        "clEnqueueWriteBuffer", 
        t1.GetElapsedTime(), 
        nBytes, 
        1);

    return SDK_SUCCESS;
}

int BufferBandwidth::timedBufCLCopy(cl_mem srcBuf, cl_mem dstBuf )
{
    CPerfCounter t1;
    cl_int status = CL_SUCCESS;
    cl_event ev;

    t1.Start();

    status = clEnqueueCopyBuffer(queue,
                              srcBuf,
                              dstBuf,
                              0,
                              0,
                              nBytes,
                              0,
                              NULL,
                              &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueCopyBuffer() failed."))
        return SDK_FAILURE;

    status = clWaitForEvents(1, &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clWaitForEvents() failed.(ev)"))
        return SDK_FAILURE;

    t1.Stop();

	// Release event
    status = clReleaseEvent(ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseEvent() failed.(ev)"))
        return SDK_FAILURE;

    tlog->Timer("%32s  %lf s %8.2lf GB/s\n", 
        "clEnqueueCopyBuffer", 
        t1.GetElapsedTime(), 
        nBytes, 
        1);

    return SDK_SUCCESS;
}

int BufferBandwidth::timedKernel(cl_kernel kernel,
                  cl_mem bufSrc,
                  cl_mem bufDst,
                  unsigned char v)
{
    cl_int status = CL_SUCCESS;
    cl_event ev = NULL;
    CPerfCounter t;

#ifdef USE_CL_PROFILER
    cl_ulong startTime, endTime;
#endif

    cl_uint nItemsPerThread = nItems / nThreads;
    size_t globalWorkSize[2] = {nThreads, 0};
    size_t localWorkSize[2] = {MAX_WAVEFRONT_SIZE, 0};

    cl_uint val = 0;

    for(int i = 0; i < sizeof(cl_uint); i++)
        val |= v << (i * 8);

    status = clSetKernelArg(kernel, 0, sizeof(void *), (void *)&bufSrc);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clSetKernelArg() failed.(bufSrc)"))
        return SDK_FAILURE;

    status = clSetKernelArg(kernel, 1, sizeof(void *), (void *)&bufDst);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clSetKernelArg() failed.(bufDst)"))
        return SDK_FAILURE;

    status = clSetKernelArg(kernel, 2, sizeof(cl_uint), (void *)&nItemsPerThread);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clSetKernelArg() failed. (nItemsPerThread)"))
        return SDK_FAILURE;

    status = clSetKernelArg(kernel, 3, sizeof(cl_uint), (void *)&val);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clSetKernelArg() failed. (val)"))
        return SDK_FAILURE;

    status = clSetKernelArg(kernel, 4, sizeof(cl_uint), (void *) &nKLoops);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clSetKernelArg() failed. (nKLoops)"))
        return SDK_FAILURE;

    t.Reset();
    t.Start();

    status = clEnqueueNDRangeKernel(queue,
                                  kernel,
                                  1,
                                  NULL,
                                  globalWorkSize,
                                  localWorkSize,
                                  0,
                                  NULL,
                                  &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clEnqueueNDRangeKernel() failed."))
        return SDK_FAILURE;

    status = clWaitForEvents(1, &ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clWaitForEvents() failed.(ev)"))
        return SDK_FAILURE;

    t.Stop();

#ifdef USE_CL_PROFILER
    status = clGetEventProfilingInfo(
                ev,
                CL_PROFILING_COMMAND_QUEUED,
                sizeof(cl_ulong),
                &startTime,
                NULL);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clGetEventProfilingInfo() failed.(CL_PROFILING_COMMAND_QUEUED)"))
        return SDK_FAILURE;

    status = clGetEventProfilingInfo(
                ev,
                CL_PROFILING_COMMAND_END,
                sizeof(cl_ulong),
                &endTime,
                NULL);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clGetEventProfilingInfo() failed.(CL_PROFILING_COMMAND_END)"))
        return SDK_FAILURE;

    tlog->Timer(
         "%32s  %lf s %8.2lf GB/s\n", 
         "clEnqueueNDRangeKernel()",
         (double)(endTime - startTime) / 1e9, 
         nBytes, 
         nKLoops);
#else
    tlog->Timer(
         "%32s  %lf s %8.2lf GB/s\n", 
         "clEnqueueNDRangeKernel()",
         t.GetElapsedTime(), 
         nBytes, 
         nKLoops);
#endif

    // Release event
    status = clReleaseEvent(ev);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseEvent() failed.(ev)"))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int BufferBandwidth::timedReadKernelVerify(
                            cl_kernel kernel,
                            cl_mem bufSrc,
                            cl_mem bufRes,
                            unsigned char v)
{
    cl_int status = timedKernel(kernel, bufSrc, bufRes, v);
    if(status == SDK_FAILURE)
        return SDK_FAILURE;

    status = clEnqueueReadBuffer(
                    queue,
                    bufRes,
                    CL_TRUE,
                    0,
                    nBytesResult,
                    memResult,
                    0,
                    NULL, 
                    NULL);
    if(!sampleCommon->checkVal(
          status,
          CL_SUCCESS,
          "clEnqueueReadBuffer() failed.(bufRes)"))
      return SDK_FAILURE;

    cl_uint sum = 0;

    for(int i = 0; i < nThreads / MAX_WAVEFRONT_SIZE; i++)
        sum += ((cl_uint*)memResult)[i];

    bool results;

    if( sum == nBytes / sizeof(cl_uint))
        results = true;
    else
        results = false;

    if(results)
        tlog->Msg("%32s\n", "verification ok");
    else
    {
        correctness = false;
        printf("Failed!\n");
        tlog->Error( "%32s\n", "verification FAILED" );
        return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}

int
BufferBandwidth::runMapTest()
{
   int nl = nLoops;
   int status = SDK_SUCCESS;

   while(nl--)
   {
      tlog->loopMarker();
      tlog->Msg("\n%s\n\n", "Host mapped write to inputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufMappedWrite(inputBuffer, nl & 0xff);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "GPU kernel read of inputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedReadKernelVerify(
                        readKernel, 
                        inputBuffer, 
                        resultBuffer, 
                        nl & 0xff);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "GPU kernel write to outputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedKernel(
                        writeKernel, 
                        resultBuffer, 
                        outputBuffer, 
                        nl & 0xff );
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "Host mapped read of outputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufMappedRead(outputBuffer, nl & 0xff);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("%s\n", "");
   }

   return SDK_SUCCESS;
}

int
BufferBandwidth::runReadWriteTest()
{
   int nl = nLoops;
   int status = SDK_SUCCESS;

   while(nl--)
   {
      tlog->loopMarker();

      tlog->Msg("\n%s\n\n", "Host CL write to inputBuffer");

      memset(memRW, nl & 0xff, nBytes);

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufCLWrite(inputBuffer, memRW);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "GPU kernel read of inputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedReadKernelVerify(
                        readKernel, 
                        inputBuffer, 
                        resultBuffer,
                        nl & 0xff );
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "GPU kernel write to outputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedKernel(
                        writeKernel, 
                        resultBuffer, 
                        outputBuffer, 
                        nl & 0xff );
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "Host CL read of outputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufCLRead(outputBuffer, memRW, nl & 0xff );
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("%s\n", "");
   }

   return SDK_SUCCESS;
}

int
BufferBandwidth::runCopyTest()
{
   int nl = nLoops;
   int status = SDK_SUCCESS;

   while(nl--)
   {
      tlog->loopMarker();

      tlog->Msg("\n%s\n\n", "Host mapped write to inputCopyBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufMappedWrite(inputCopyBuffer, nl & 0xff);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "CL copy of inputCopyBuffer to inputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufCLCopy(inputCopyBuffer, inputBuffer);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "GPU kernel read of inputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedReadKernelVerify(
                        readKernel, 
                        inputBuffer, 
                        resultBuffer, 
                        nl & 0xff);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "GPU kernel write to outputBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedKernel(
                        writeKernel, 
                        resultBuffer, 
                        outputBuffer, 
                        nl & 0xff );
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "CL copy of outputBuffer to outputCopyBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufCLCopy(outputBuffer, outputCopyBuffer);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("\n%s\n\n", "Host mapped read of outputCopyBuffer");

      for(int i = 0; i < nRepeats; i++)
      {
          status = timedBufMappedRead(outputCopyBuffer, nl & 0xff);
          if(status == SDK_FAILURE)
              return status;
      }

      tlog->Msg("%s\n", "");
   }

   return SDK_SUCCESS;
}


int 
BufferBandwidth::run()
{
    int status = SDK_SUCCESS;

    if(verify)
    {
        status = runMapTest(); 
        if(status == SDK_FAILURE)
            return status;
        
        status = runReadWriteTest(); 
        if(status == SDK_FAILURE)
            return status;

        status = runCopyTest(); 
        if(status == SDK_FAILURE)
            return status;

        if(correctness)
        {
            printf("Passed!!\n");
            return SDK_SUCCESS;
        }
        else
        {
            printf("Failed\n");
            return SDK_FAILURE;
        }

    }
    else
    {
        // Run specified test
        switch(whichTest)
        {
           case 0: 
                status = runMapTest(); 
                if(status == SDK_FAILURE)
                    return status;
                break;
           case 1: 
                status = runReadWriteTest(); 
                if(status == SDK_FAILURE)
                    return status;
                break;
           case 2: 
                status = runCopyTest(); 
                if(status == SDK_FAILURE)
                    return status;
                break;
        }
        // Run Host memory performance test.
        if(!doHost)
            assessHostMemPerf(memIn, memOut, nBytes, quiet);
    }

    return SDK_SUCCESS;
}

void 
BufferBandwidth::printStats()
{
    if(!quiet && !verify)
    {
#ifdef _WIN32
        printf( "Build:               _WINxx" ); 
#ifdef _DEBUG
        printf( " DEBUG");
#else
        printf( " release");
#endif
        printf( "\n" );
#else
#ifdef NDEBUG
        printf("Build:               release\n");
#else
        printf("Build:               DEBUG\n");
#endif
#endif

        printf("GPU work items:      %d\n", nThreads);
        printf("Buffer size:         %d\n", nBytes);
        printf("Timing loops:        %d\n", nLoops);
        printf("Repeats:             %d\n", nRepeats);
        printf("Kernel loops:        %d\n", nKLoops);
        printf("Wavefronts/SIMD:     %d\n", numWavefronts);

        printf("inputBuffer:         ");

        for(int i = 0; i < sizeof(memFlags) / sizeof(memFlags[0]); i++)
            if(inFlags & memFlags[i].f)
                printf("%s ", memFlags[i].s);

        printf("\noutputBuffer:        ");

        for(int i = 0; i < sizeof(memFlags) / sizeof(memFlags[0]); i++)
            if(outFlags & memFlags[i].f)
                printf("%s ", memFlags[i].s);

        if(whichTest == 2)
        {
            printf("\n*CopyBuffer:         ");

            for(int i = 0; i < sizeof(memFlags) / sizeof(memFlags[0]); i++)
                if(copyFlags & memFlags[i].f)
                    printf("%s ", memFlags[i].s);
        }

        printf("\n");

        if(printLog) 
        {
            printf("\n\nLOOP ITERATIONS\n");
            printf("---------------\n");
            tlog->printLog();
        }

        printf(
            "\nAVERAGES (over loops %d - %d, use -l for complete log)\n", 
            nSkip, 
            nLoops - 1);
        printf("--------\n");

        tlog->printSummary(nSkip);

        printf("\n");
    }
}

int 
BufferBandwidth::verifyResults()
{
    return SDK_SUCCESS;
}

int 
BufferBandwidth::cleanup()
{
    /* Releases OpenCL resources (Context, Memory etc.) */
    cl_int status;

    if(inputBuffer)
    {
        status = clReleaseMemObject(inputBuffer);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(outputBuffer)
    {
        status = clReleaseMemObject(outputBuffer);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(resultBuffer)
    {
        status = clReleaseMemObject(resultBuffer);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(inputCopyBuffer)
    {
        status = clReleaseMemObject(inputCopyBuffer);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(outputCopyBuffer)
    {
        status = clReleaseMemObject(outputCopyBuffer);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    status = clReleaseKernel(readKernel);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseProgram failed."))
        return SDK_FAILURE;

    status = clReleaseKernel(writeKernel);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseProgram failed."))
        return SDK_FAILURE;

    status = clReleaseProgram(program);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseProgram failed."))
        return SDK_FAILURE;
 
    status = clReleaseCommandQueue(queue);
     if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseCommandQueue failed."))
        return SDK_FAILURE;

    status = clReleaseContext(context);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseContext failed."))
        return SDK_FAILURE;

    if(memIn) 
    {
        #ifdef _WIN32
            _aligned_free(memIn);
        #else
            free(memIn);
        #endif
        memIn = NULL;
    }

    if(memOut) 
    {
        #ifdef _WIN32
            _aligned_free(memOut);
        #else
            free(memOut);
        #endif
        memOut = NULL;
    }

    if(memResult) 
    {
        #ifdef _WIN32
            _aligned_free(memResult);
        #else
            free(memResult);
        #endif
        memResult = NULL;
    }

    if(memRW) 
    {
        #ifdef _WIN32
            _aligned_free(memRW);
        #else
            free(memRW);
        #endif
        memRW = NULL;
    }

    delete tlog;

    /* release device list */
    if(devices)
        free(devices);

    return SDK_SUCCESS;
}

int 
main(int argc, char * argv[])
{
    BufferBandwidth clBufferBandwidth("OpenCL Constant Bandwidth");

    clBufferBandwidth.initialize();
    if(clBufferBandwidth.parseExtraCommandLineOptions(argc, argv))
        return SDK_FAILURE;


    if(clBufferBandwidth.isDumpBinaryEnabled())
    {
        return clBufferBandwidth.genBinaryImage();
    }
    else
    {
        int state = clBufferBandwidth.setup();
        if(state != SDK_SUCCESS)
        {
            if(state == SDK_EXPECTED_FAILURE)
                return SDK_SUCCESS;
            else
                return SDK_FAILURE;
        }

        state = clBufferBandwidth.run();
        if(state == SDK_FAILURE)
          return SDK_FAILURE;

        clBufferBandwidth.printStats();
        if(clBufferBandwidth.cleanup() == SDK_FAILURE)
          return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}
