/* ============================================================

Copyright (c) 2011 Advanced Micro Devices, Inc.  All rights reserved.
 
Redistribution and use of this material is permitted under the following 
conditions:
 
Redistributions must retain the above copyright notice and all terms of this 
license.
 
In no event shall anyone redistributing or accessing or using this material 
commence or participate in any arbitration or legal action relating to this 
material against Advanced Micro Devices, Inc. or any copyright holders or 
contributors. The foregoing shall survive any expiration or termination of 
this license or any agreement or access or use related to this material. 

ANY BREACH OF ANY TERM OF THIS LICENSE SHALL RESULT IN THE IMMEDIATE REVOCATION 
OF ALL RIGHTS TO REDISTRIBUTE, ACCESS OR USE THIS MATERIAL.

THIS MATERIAL IS PROVIDED BY ADVANCED MICRO DEVICES, INC. AND ANY COPYRIGHT 
HOLDERS AND CONTRIBUTORS "AS IS" IN ITS CURRENT CONDITION AND WITHOUT ANY 
REPRESENTATIONS, GUARANTEE, OR WARRANTY OF ANY KIND OR IN ANY WAY RELATED TO 
SUPPORT, INDEMNITY, ERROR FREE OR UNINTERRUPTED OPERA TION, OR THAT IT IS FREE 
FROM DEFECTS OR VIRUSES.  ALL OBLIGATIONS ARE HEREBY DISCLAIMED - WHETHER 
EXPRESS, IMPLIED, OR STATUTORY - INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED 
WARRANTIES OF TITLE, MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, 
ACCURACY, COMPLETENESS, OPERABILITY, QUALITY OF SERVICE, OR NON-INFRINGEMENT. 
IN NO EVENT SHALL ADVANCED MICRO DEVICES, INC. OR ANY COPYRIGHT HOLDERS OR 
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, PUNITIVE,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, REVENUE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED OR BASED ON ANY THEORY OF LIABILITY 
ARISING IN ANY WAY RELATED TO THIS MATERIAL, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE. THE ENTIRE AND AGGREGATE LIABILITY OF ADVANCED MICRO DEVICES, 
INC. AND ANY COPYRIGHT HOLDERS AND CONTRIBUTORS SHALL NOT EXCEED TEN DOLLARS 
(US $10.00). ANYONE REDISTRIBUTING OR ACCESSING OR USING THIS MATERIAL ACCEPTS 
THIS ALLOCATION OF RISK AND AGREES TO RELEASE ADVANCED MICRO DEVICES, INC. AND 
ANY COPYRIGHT HOLDERS AND CONTRIBUTORS FROM ANY AND ALL LIABILITIES, 
OBLIGATIONS, CLAIMS, OR DEMANDS IN EXCESS OF TEN DOLLARS (US $10.00). THE 
FOREGOING ARE ESSENTIAL TERMS OF THIS LICENSE AND, IF ANY OF THESE TERMS ARE 
CONSTRUED AS UNENFORCEABLE, FAIL IN ESSENTIAL PURPOSE, OR BECOME VOID OR 
DETRIMENTAL TO ADVANCED MICRO DEVICES, INC. OR ANY COPYRIGHT HOLDERS OR 
CONTRIBUTORS FOR ANY REASON, THEN ALL RIGHTS TO REDISTRIBUTE, ACCESS OR USE 
THIS MATERIAL SHALL TERMINATE IMMEDIATELY. MOREOVER, THE FOREGOING SHALL 
SURVIVE ANY EXPIRATION OR TERMINATION OF THIS LICENSE OR ANY AGREEMENT OR 
ACCESS OR USE RELATED TO THIS MATERIAL.

NOTICE IS HEREBY PROVIDED, AND BY REDISTRIBUTING OR ACCESSING OR USING THIS 
MATERIAL SUCH NOTICE IS ACKNOWLEDGED, THAT THIS MATERIAL MAY BE SUBJECT TO 
RESTRICTIONS UNDER THE LAWS AND REGULATIONS OF THE UNITED STATES OR OTHER 
COUNTRIES, WHICH INCLUDE BUT ARE NOT LIMITED TO, U.S. EXPORT CONTROL LAWS SUCH 
AS THE EXPORT ADMINISTRATION REGULATIONS AND NATIONAL SECURITY CONTROLS AS 
DEFINED THEREUNDER, AS WELL AS STATE DEPARTMENT CONTROLS UNDER THE U.S. 
MUNITIONS LIST. THIS MATERIAL MAY NOT BE USED, RELEASED, TRANSFERRED, IMPORTED,
EXPORTED AND/OR RE-EXPORTED IN ANY MANNER PROHIBITED UNDER ANY APPLICABLE LAWS, 
INCLUDING U.S. EXPORT CONTROL LAWS REGARDING SPECIFICALLY DESIGNATED PERSONS, 
COUNTRIES AND NATIONALS OF COUNTRIES SUBJECT TO NATIONAL SECURITY CONTROLS. 
MOREOVER, THE FOREGOING SHALL SURVIVE ANY EXPIRATION OR TERMINATION OF ANY 
LICENSE OR AGREEMENT OR ACCESS OR USE RELATED TO THIS MATERIAL.

NOTICE REGARDING THE U.S. GOVERNMENT AND DOD AGENCIES: This material is 
provided with "RESTRICTED RIGHTS" and/or "LIMITED RIGHTS" as applicable to 
computer software and technical data, respectively. Use, duplication, 
distribution or disclosure by the U.S. Government and/or DOD agencies is 
subject to the full extent of restrictions in all applicable regulations, 
including those found at FAR52.227 and DFARS252.227 et seq. and any successor 
regulations thereof. Use of this material by the U.S. Government and/or DOD 
agencies is acknowledgment of the proprietary rights of any copyright holders 
and contributors, including those of Advanced Micro Devices, Inc., as well as 
the provisions of FAR52.227-14 through 23 regarding privately developed and/or 
commercial computer software.

This license forms the entire agreement regarding the subject matter hereof and 
supersedes all proposals and prior discussions and writings between the parties 
with respect thereto. This license does not affect any ownership, rights, title,
or interest in, or relating to, this material. No terms of this license can be 
modified or waived, and no breach of this license can be excused, unless done 
so in a writing signed by all affected parties. Each term of this license is 
separately enforceable. If any term of this license is determined to be or 
becomes unenforceable or illegal, such term shall be reformed to the minimum 
extent necessary in order for this license to remain in effect in accordance 
with its terms as modified by such reformation. This license shall be governed 
by and construed in accordance with the laws of the State of Texas without 
regard to rules on conflicts of law of any state or jurisdiction or the United 
Nations Convention on the International Sale of Goods. All disputes arising out 
of this license shall be subject to the jurisdiction of the federal and state 
courts in Austin, Texas, and all defenses are hereby waived concerning personal 
jurisdiction and venue of these courts.

============================================================ */

#include "TransferOverlap.hpp"
#include <string>


struct _memFlags {cl_mem_flags f;
                  const char *s;} 
                  memFlags[] = 
                  {
                     { CL_MEM_READ_ONLY,              "CL_MEM_READ_ONLY" },
                     { CL_MEM_WRITE_ONLY,             "CL_MEM_WRITE_ONLY" },
                     { CL_MEM_READ_WRITE,             "CL_MEM_READ_WRITE" },
                     { CL_MEM_ALLOC_HOST_PTR,         "CL_MEM_ALLOC_HOST_PTR" },
                     { CL_MEM_USE_PERSISTENT_MEM_AMD, "CL_MEM_USE_PERSISTENT_MEM_AMD"}
                  };

int nFlags = sizeof(memFlags) / sizeof(memFlags[0]);


int 
TransferOverlap::setupTransferOverlap()
{

    // Increase the input valuse if -e option is not used
    if(!verify)
    {
        nLoops = 50;
        nSkip =  3;
        nKLoops = 45;
        nBytes = 16 * 1024 * 1024; 
    }

    timeLog = new TestLog(nLoops * 50);
    if(timeLog == NULL)
    { 
        sampleCommon->error("Failed to allocate host memory. (timeLog)");
        return SDK_FAILURE;
    }

    // Educated guess of optimal work size
    int minBytes = MAX_WAVEFRONT_SIZE * sizeof(cl_uint) * 4;

    nBytes = (nBytes / minBytes) * minBytes;
    nBytes = nBytes < minBytes ? minBytes : nBytes;
    nItems = nBytes / (4 * sizeof(cl_uint));
    
    int maxThreads = nBytes / (4 * sizeof( cl_uint ));

    nThreads = deviceMaxComputeUnits * numWavefronts * MAX_WAVEFRONT_SIZE;
    
    if(nThreads > maxThreads)
        nThreads = maxThreads;
    else
    {
        while(nItems % nThreads != 0)
            nThreads += MAX_WAVEFRONT_SIZE;
    }

    nBytesResult = (nThreads / MAX_WAVEFRONT_SIZE) * sizeof(cl_uint) * 2;

    return SDK_SUCCESS;
}

int 
TransferOverlap::genBinaryImage()
{
    cl_int status = CL_SUCCESS;

    /*
     * Have a look at the available platforms and pick either
     * the AMD one if available or a reasonable default.
     */
    cl_uint numPlatforms;
    cl_platform_id platform = NULL;
    status = clGetPlatformIDs(0, NULL, &numPlatforms);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetPlatformIDs failed."))
        return SDK_FAILURE;

    if (0 < numPlatforms) 
    {
        cl_platform_id* platforms = new cl_platform_id[numPlatforms];
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clGetPlatformIDs failed."))
            return SDK_FAILURE;

        char platformName[100];
        for (unsigned i = 0; i < numPlatforms; ++i) 
        {
            status = clGetPlatformInfo(platforms[i],
                                       CL_PLATFORM_VENDOR,
                                       sizeof(platformName),
                                       platformName,
                                       NULL);

            if(!sampleCommon->checkVal(status,
                                       CL_SUCCESS,
                                       "clGetPlatformInfo failed."))
                return SDK_FAILURE;

            platform = platforms[i];
            if (!strcmp(platformName, "Advanced Micro Devices, Inc.")) 
                break;
        }
        std::cout << "Platform found : " << platformName << "\n";
        delete[] platforms;
    }

    if(NULL == platform)
    {
        sampleCommon->error("NULL platform found so Exiting Application.");
        return SDK_FAILURE;
    }

    /*
     * If we could find our platform, use it. Otherwise use just available platform.
     */
    cl_context_properties cps[5] = 
    {
        CL_CONTEXT_PLATFORM, 
        (cl_context_properties)platform, 
        CL_CONTEXT_OFFLINE_DEVICES_AMD,
        (cl_context_properties)1,
        0
    };

    context = clCreateContextFromType(cps,
                                      CL_DEVICE_TYPE_ALL,
                                      NULL,
                                      NULL,
                                      &status);

    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clCreateContextFromType failed."))
        return SDK_FAILURE;

    /* create a CL program using the kernel source */
    streamsdk::SDKFile kernelFile;
    std::string kernelPath = sampleCommon->getPath();
    kernelPath.append("TransferOverlap_Kernels.cl");
    if(!kernelFile.open(kernelPath.c_str()))
    {
        std::cout << "Failed to load kernel file : " << kernelPath << std::endl;
        return SDK_FAILURE;
    }
    const char * source = kernelFile.source().c_str();
    size_t sourceSize[] = {strlen(source)};
    program = clCreateProgramWithSource(context,
                                        1,
                                        &source,
                                        sourceSize,
                                        &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clCreateProgramWithSource failed."))
        return SDK_FAILURE;
    
    std::string flagsStr = std::string("");

    // Get additional options
    if(isComplierFlagsSpecified())
    {
        streamsdk::SDKFile flagsFile;
        std::string flagsPath = sampleCommon->getPath();
        flagsPath.append(flags.c_str());
        if(!flagsFile.open(flagsPath.c_str()))
        {
            std::cout << "Failed to load flags file: " << flagsPath << std::endl;
            return SDK_FAILURE;
        }
        flagsFile.replaceNewlineWithSpaces();
        const char * flags = flagsFile.source().c_str();
        flagsStr.append(flags);
    }

    if(flagsStr.size() != 0)
        std::cout << "Build Options are : " << flagsStr.c_str() << std::endl;

    /* create a cl program executable for all the devices specified */
    status = clBuildProgram(program,
                            0,
                            NULL,
                            flagsStr.c_str(),
                            NULL,
                            NULL);

    sampleCommon->checkVal(status,
                        CL_SUCCESS,
                        "clBuildProgram failed.");

    size_t numDevices;
    status = clGetProgramInfo(program, 
                           CL_PROGRAM_NUM_DEVICES,
                           sizeof(numDevices),
                           &numDevices,
                           NULL );
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_NUM_DEVICES) failed."))
        return SDK_FAILURE;

    std::cout << "Number of devices found : " << numDevices << "\n\n";
    devices = (cl_device_id *)malloc( sizeof(cl_device_id) * numDevices );
    if(devices == NULL)
    {
        sampleCommon->error("Failed to allocate host memory.(devices)");
        return SDK_FAILURE;
    }
    /* grab the handles to all of the devices in the program. */
    status = clGetProgramInfo(program, 
                              CL_PROGRAM_DEVICES, 
                              sizeof(cl_device_id) * numDevices,
                              devices,
                              NULL );
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_DEVICES) failed."))
        return SDK_FAILURE;

    /* figure out the sizes of each of the binaries. */
    size_t *binarySizes = (size_t*)malloc( sizeof(size_t) * numDevices );
    if(devices == NULL)
    {
        sampleCommon->error("Failed to allocate host memory.(binarySizes)");
        return SDK_FAILURE;
    }
    
    status = clGetProgramInfo(program, 
                              CL_PROGRAM_BINARY_SIZES,
                              sizeof(size_t) * numDevices, 
                              binarySizes, NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_BINARY_SIZES) failed."))
        return SDK_FAILURE;

    size_t i = 0;
    /* copy over all of the generated binaries. */
    char **binaries = (char **)malloc( sizeof(char *) * numDevices );
    if(binaries == NULL)
    {
        sampleCommon->error("Failed to allocate host memory.(binaries)");
        return SDK_FAILURE;
    }

    for(i = 0; i < numDevices; i++)
    {
        if(binarySizes[i] != 0)
        {
            binaries[i] = (char *)malloc( sizeof(char) * binarySizes[i]);
            if(binaries[i] == NULL)
            {
                sampleCommon->error("Failed to allocate host memory.(binaries[i])");
                return SDK_FAILURE;
            }
        }
        else
        {
            binaries[i] = NULL;
        }
    }
    status = clGetProgramInfo(program, 
                              CL_PROGRAM_BINARIES,
                              sizeof(char *) * numDevices, 
                              binaries, 
                              NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetProgramInfo(CL_PROGRAM_BINARIES) failed."))
        return SDK_FAILURE;

    /* dump out each binary into its own separate file. */
    for(i = 0; i < numDevices; i++)
    {
        char fileName[100];
        sprintf(fileName, "%s.%d", dumpBinary.c_str(), (int)i);
        if(binarySizes[i] != 0)
        {
            char deviceName[1024];
            status = clGetDeviceInfo(devices[i], 
                                     CL_DEVICE_NAME, 
                                     sizeof(deviceName),
                                     deviceName, 
                                     NULL);
            if(!sampleCommon->checkVal(status,
                                       CL_SUCCESS,
                                       "clGetDeviceInfo(CL_DEVICE_NAME) failed."))
                return SDK_FAILURE;

            printf( "%s binary kernel: %s\n", deviceName, fileName);
            streamsdk::SDKFile BinaryFile;
            if(!BinaryFile.writeBinaryToFile(fileName, 
                                             binaries[i], 
                                             binarySizes[i]))
            {
                std::cout << "Failed to load kernel file : " << fileName << std::endl;
                return SDK_FAILURE;
            }
        }
        else
        {
            printf("Skipping %s since there is no binary data to write!\n",
                    fileName);
        }
    }

    // Release all resouces and memory
    for(i = 0; i < numDevices; i++)
    {
        if(binaries[i] != NULL)
        {
            free(binaries[i]);
            binaries[i] = NULL;
        }
    }

    if(binaries != NULL)
    {
        free(binaries);
        binaries = NULL;
    }

    if(binarySizes != NULL)
    {
        free(binarySizes);
        binarySizes = NULL;
    }

    if(devices != NULL)
    {
        free(devices);
        devices = NULL;
    }

    status = clReleaseProgram(program);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clReleaseProgram failed."))
        return SDK_FAILURE;

    status = clReleaseContext(context);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clReleaseContext failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}


int
TransferOverlap::setupCL(void)
{
    cl_int status = 0;
    size_t deviceListSize;

    cl_device_type dType;
    
    if(deviceType.compare("cpu") == 0)
    {
        dType = CL_DEVICE_TYPE_CPU;
    }
    else //deviceType = "gpu" 
    {
        dType = CL_DEVICE_TYPE_GPU;
        if(isThereGPU() == false)
        {
            std::cout << "GPU not found. Falling back to CPU device" << std::endl;
            dType = CL_DEVICE_TYPE_CPU;
        }
    }

    /*
     * Have a look at the available platforms and pick either
     * the AMD one if available or a reasonable default.
     */

    cl_uint numPlatforms;
    cl_platform_id platform = NULL;
    status = clGetPlatformIDs(0, NULL, &numPlatforms);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetPlatformIDs failed."))
    {
        return SDK_FAILURE;
    }
    if (0 < numPlatforms) 
    {
        cl_platform_id* platforms = new cl_platform_id[numPlatforms];
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clGetPlatformIDs failed."))
        {
            return SDK_FAILURE;
        }

        if(isPlatformEnabled())
        {
            platform = platforms[platformId];
        }
        else
        {
            for (unsigned i = 0; i < numPlatforms; ++i) 
            {
                char pbuf[100];
                status = clGetPlatformInfo(platforms[i],
                                           CL_PLATFORM_VENDOR,
                                           sizeof(pbuf),
                                           pbuf,
                                           NULL);

                if(!sampleCommon->checkVal(status,
                                           CL_SUCCESS,
                                           "clGetPlatformInfo failed."))
                {
                    return SDK_FAILURE;
                }

                platform = platforms[i];
                if (!strcmp(pbuf, "Advanced Micro Devices, Inc.")) 
                {
                    break;
                }
            }
        }
        delete[] platforms;
    }

    if(NULL == platform)
    {
        sampleCommon->error("NULL platform found so Exiting Application.");
        return SDK_FAILURE;
    }

    // Display available devices.
    if(!sampleCommon->displayDevices(platform, dType))
    {
        sampleCommon->error("sampleCommon::displayDevices() failed");
        return SDK_FAILURE;
    }

    /*
     * If we could find our platform, use it. Otherwise use just available platform.
     */

    cl_context_properties cps[3] = 
    {
        CL_CONTEXT_PLATFORM, 
        (cl_context_properties)platform, 
        0
    };

    context = clCreateContextFromType(cps,
                                      dType,
                                      NULL,
                                      NULL,
                                      &status);
    if(!sampleCommon->checkVal(status, 
                               CL_SUCCESS,
                               "clCreateContextFromType failed."))
        return SDK_FAILURE;

    /* First, get the size of device list data */
    status = clGetContextInfo(context, 
                              CL_CONTEXT_DEVICES, 
                              0, 
                              NULL, 
                              &deviceListSize);
    if(!sampleCommon->checkVal(status, 
                               CL_SUCCESS,
                               "clGetContextInfo failed."))
        return SDK_FAILURE;

    int deviceCount = (int)(deviceListSize / sizeof(cl_device_id));
    if(!sampleCommon->validateDeviceId(deviceId, deviceCount))
    {
        sampleCommon->error("sampleCommon::validateDeviceId() failed");
        return SDK_FAILURE;
    }

    /* Now allocate memory for device list based on the size we got earlier */
    devices = (cl_device_id*)malloc(deviceListSize);
    if(devices == NULL) 
    {
        sampleCommon->error("Failed to allocate memory (devices).");
        return SDK_FAILURE;
    }

    /* Now, get the device list data */
    status = clGetContextInfo(context, 
                              CL_CONTEXT_DEVICES, 
                              deviceListSize, 
                              devices, 
                              NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clGetGetContextInfo failed."))
        return SDK_FAILURE;


    status = clGetDeviceInfo(devices[deviceId],
                           CL_DEVICE_MAX_COMPUTE_UNITS,
                           sizeof(cl_uint),
                           &deviceMaxComputeUnits,
                           NULL);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clGetDeviceInfo CL_DEVICE_MAX_COMPUTE_UNITS failed."))
        return SDK_FAILURE;

    // Check device extensions
    char* deviceExtensions = NULL;;
    size_t extStringSize = 0;

    // Get device extensions 
    status = clGetDeviceInfo(devices[deviceId], 
                            CL_DEVICE_EXTENSIONS,
                            0, 
                            NULL, 
                            &extStringSize);
    if(!sampleCommon->checkVal(status,
                                CL_SUCCESS, 
                                "clGetDeviceInfo failed.(CL_DEVICE_EXTENSIONS)"))
        return SDK_FAILURE;
    
    deviceExtensions = new char[extStringSize];
    if(NULL == deviceExtensions)
    {
        sampleCommon->error("Failed to allocate memory(deviceExtensions)");
        return  SDK_FAILURE;
    }

    status = clGetDeviceInfo(devices[deviceId],
                            CL_DEVICE_EXTENSIONS, 
                            extStringSize,
                            deviceExtensions,
                            NULL);
    if(!sampleCommon->checkVal(status,
                                CL_SUCCESS, 
                                "clGetDeviceInfo failed.(CL_DEVICE_EXTENSIONS)"))
        return SDK_FAILURE;


    std::string buildOptions = std::string("");
    /* Check if cl_khr_fp64 extension is supported */
    if(!strstr(deviceExtensions, "cl_khr_local_int32_base_atomics"))
    {
        sampleCommon->expectedError("Device does not support cl_khr_local_int32_base_atomics extension!");
        return SDK_EXPECTED_FAILURE;
    }
    delete deviceExtensions;

    // Get OpenCL device version
    char deviceVersion[1024];
    status = clGetDeviceInfo(devices[deviceId], 
        CL_DEVICE_VERSION, 
        sizeof(deviceVersion), 
        deviceVersion, 
        NULL);

    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS, 
        "clGetDeviceInfo failed.(CL_DEVICE_VERSION)"))
        return SDK_FAILURE;

    std::string deviceVersionStr = std::string(deviceVersion);
    size_t vStart = deviceVersionStr.find(" ", 0);
    size_t vEnd = deviceVersionStr.find(" ", vStart + 1);
    std::string vStrVal = deviceVersionStr.substr(vStart + 1, vEnd - vStart - 1);

    // Check of OPENCL_C_VERSION if device version is 1.1 or later
    if(vStrVal.compare("1.0") > 0)
    {
        //Get OPENCL_C_VERSION 
        status = clGetDeviceInfo(devices[deviceId], 
            CL_DEVICE_OPENCL_C_VERSION, 
            sizeof(deviceVersion), 
            deviceVersion, 
            NULL);

        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clGetDeviceInfo failed.(CL_DEVICE_OPENCL_C_VERSION)"))
            return SDK_FAILURE;

        // Exit if OpenCL C device version is 1.0
        deviceVersionStr = std::string(deviceVersion);
        vStart = deviceVersionStr.find(" ", 0);
        vEnd = deviceVersionStr.find(" ", vStart + 1);
        vStrVal = deviceVersionStr.substr(vStart + 1, vEnd - vStart - 1);
        if(vStrVal.compare("1.0") <= 0)
        {
            sampleCommon->expectedError("Unsupported device! Required CL_DEVICE_OPENCL_C_VERSION as 1.1");
            return SDK_EXPECTED_FAILURE;
        }
    }
    else
    {
        sampleCommon->expectedError("Unsupported device! Required CL_DEVICE_OPENCL_C_VERSION as 1.1");
        return SDK_EXPECTED_FAILURE;
    }

    //Setup application data
    if(setupTransferOverlap() != SDK_SUCCESS)
        return SDK_FAILURE;

    {
        /* The block is to move the declaration of prop closer to its use */
        cl_command_queue_properties prop = 0;

        queue = clCreateCommandQueue(context, 
                                            devices[deviceId], 
                                            prop, 
                                            &status);
        if(!sampleCommon->checkVal(status,
                                   0,
                                   "clCreateCommandQueue failed."))
            return SDK_FAILURE;
    }

    inputBuffer1 = clCreateBuffer(context, inFlags, nBytes, NULL, &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(inputBuffer1)"))
        return SDK_FAILURE;

    inputBuffer2 = clCreateBuffer(context, inFlags, nBytes, NULL, &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(inputBuffer2)"))
        return SDK_FAILURE;

    resultBuffer1 = clCreateBuffer(context,
                                   CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR,
                                   nBytesResult,
                                   NULL, 
                                   &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(resultBuffer1)"))
        return SDK_FAILURE;

    resultBuffer2 = clCreateBuffer(context,
                                   CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR,
                                   nBytesResult,
                                   NULL, 
                                   &status);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS, 
                               "clCreateBuffer() failed.(resultBuffer2)"))
        return SDK_FAILURE;
    
    /* create a CL program using the kernel source */
    streamsdk::SDKFile kernelFile;
    std::string kernelPath = sampleCommon->getPath();

    if(isLoadBinaryEnabled())
    {
        kernelPath.append(loadBinary.c_str());
        if(!kernelFile.readBinaryFromFile(kernelPath.c_str()))
        {
            std::cout << "Failed to load kernel file : " << kernelPath << std::endl;
            return SDK_FAILURE;
        }

        const char * binary = kernelFile.source().c_str();
        size_t binarySize = kernelFile.source().size();
        program = clCreateProgramWithBinary(context,
                                            1,
                                            &devices[deviceId], 
                                            (const size_t *)&binarySize,
                                            (const unsigned char**)&binary,
                                            NULL,
                                            &status);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clCreateProgramWithBinary failed."))
        {
            return SDK_FAILURE;
        }

    }
    else
    {
        kernelPath.append("TransferOverlap_Kernels.cl");
        if(!kernelFile.open(kernelPath.c_str()))
        {
            std::cout << "Failed to load kernel file: " << kernelPath << std::endl;
            return SDK_FAILURE;
        }
        const char * source = kernelFile.source().c_str();
        size_t sourceSize[] = {strlen(source)};
        program = clCreateProgramWithSource(context,
                                            1,
                                            &source,
                                            sourceSize,
                                            &status);
        if(!sampleCommon->checkVal(status,
                                   CL_SUCCESS,
                                   "clCreateProgramWithSource failed."))
            return SDK_FAILURE;
    }

    std::string flagsStr = std::string("");

    // Get additional options
    if(isComplierFlagsSpecified())
    {
        streamsdk::SDKFile flagsFile;
        std::string flagsPath = sampleCommon->getPath();
        flagsPath.append(flags.c_str());
        if(!flagsFile.open(flagsPath.c_str()))
        {
            std::cout << "Failed to load flags file: " << flagsPath << std::endl;
            return SDK_FAILURE;
        }
        flagsFile.replaceNewlineWithSpaces();
        const char * flags = flagsFile.source().c_str();
        flagsStr.append(flags);
    }

    if(flagsStr.size() != 0)
        std::cout << "Build Options are : " << flagsStr.c_str() << std::endl;

    

    /* create a cl program executable for all the devices specified */
    status = clBuildProgram(program, 1, &devices[deviceId], flagsStr.c_str(), NULL, NULL);
    if(status != CL_SUCCESS)
    {
        if(status == CL_BUILD_PROGRAM_FAILURE)
        {
            cl_int logStatus;
            char *buildLog = NULL;
            size_t buildLogSize = 0;
            logStatus = clGetProgramBuildInfo (program, 
                devices[deviceId], 
                CL_PROGRAM_BUILD_LOG, 
                buildLogSize, 
                buildLog, 
                &buildLogSize);
            if(!sampleCommon->checkVal(
                logStatus,
                CL_SUCCESS,
                "clGetProgramBuildInfo failed."))
                return SDK_FAILURE;

            buildLog = (char*)malloc(buildLogSize);
            if(buildLog == NULL)
            {
                sampleCommon->error("Failed to allocate host memory. (buildLog)");
                return SDK_FAILURE;
            }
            memset(buildLog, 0, buildLogSize);

            logStatus = clGetProgramBuildInfo (program, 
                devices[deviceId], 
                CL_PROGRAM_BUILD_LOG, 
                buildLogSize, 
                buildLog, 
                NULL);
            if(!sampleCommon->checkVal(
                logStatus,
                CL_SUCCESS,
                "clGetProgramBuildInfo failed."))
            {
                free(buildLog);
                return SDK_FAILURE;
            }

            std::cout << " \n\t\t\tBUILD LOG\n";
            std::cout << " ************************************************\n";
            std::cout << buildLog << std::endl;
            std::cout << " ************************************************\n";
            free(buildLog);
        }

        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clBuildProgram failed."))
            return SDK_FAILURE;
    }

    /* ConstantBuffer bandwidth from single access */
    readKernel = clCreateKernel(program, "readKernel", &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clCreateKernel failed.(readKernel)"))
        return SDK_FAILURE;

    size_t kernelWorkGroupSize = 0;

    /* Check whether specified local group size is possible on current kernel */
    status = clGetKernelWorkGroupInfo(readKernel,
                                      devices[deviceId],
                                      CL_KERNEL_WORK_GROUP_SIZE,
                                      sizeof(size_t),
                                      &kernelWorkGroupSize,
                                      0);
    if(!sampleCommon->checkVal(status,
                               CL_SUCCESS,
                               "clGetKernelWorkGroupInfo failed."))
        return SDK_FAILURE;


    // If local groupSize exceeds the maximum supported on kernel  fall back
    if(MAX_WAVEFRONT_SIZE > kernelWorkGroupSize)
    {
        if(!quiet)
        {
            std::cout << "Out of Resources!" << std::endl;
            std::cout << "Group Size specified : " << MAX_WAVEFRONT_SIZE << std::endl;
            std::cout << "Max Group Size supported on the kernel(readKernel) : " 
                      << kernelWorkGroupSize << std::endl;
            std::cout << "Falling back to " << kernelWorkGroupSize << std::endl;
        }
        return SDK_FAILURE;
    }

    writeKernel = clCreateKernel(program, "writeKernel", &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clCreateKernel failed.(writeKernel)"))
        return SDK_FAILURE;

    // Check whether specified local group size is possible on current kernel
    status = clGetKernelWorkGroupInfo(writeKernel,
                                      devices[deviceId],
                                      CL_KERNEL_WORK_GROUP_SIZE,
                                      sizeof(size_t),
                                      &kernelWorkGroupSize,
                                      0);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clGetKernelWorkGroupInfo failed."))
        return SDK_FAILURE;


    // If local group size exceeds the maximum supported on kernel fall back 
    if(MAX_WAVEFRONT_SIZE > kernelWorkGroupSize)
    {
        if(!quiet)
        {
            std::cout << "Out of Resources!" << std::endl;
            std::cout << "Group Size specified : " << MAX_WAVEFRONT_SIZE << std::endl;
            std::cout << "Max Group Size supported on the kernel(writeKernel) : " 
                      << kernelWorkGroupSize << std::endl;
            std::cout << "Falling back to " << kernelWorkGroupSize << std::endl;
        }
        return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}

int 
TransferOverlap::initialize()
{
    /* Call base class Initialize to get default configuration */
    if(!this->SDKSample::initialize())
        return SDK_FAILURE;

    streamsdk::Option* numLoops = new streamsdk::Option;

    numLoops->_sVersion = "i";
    numLoops->_lVersion = "iterations";
    numLoops->_description = "Number of timing loops";
    numLoops->_type = streamsdk::CA_ARG_INT;
    numLoops->_value = &nLoops;

    sampleArgs->AddOption(numLoops);
    delete numLoops;

    streamsdk::Option* skipLoops = new streamsdk::Option;

    skipLoops->_sVersion = "s";
    skipLoops->_lVersion = "skip";
    skipLoops->_description = "skip first n iterations for average";
    skipLoops->_type = streamsdk::CA_ARG_INT;
    skipLoops->_value = &nSkip;

    sampleArgs->AddOption(skipLoops);
    delete skipLoops;


    streamsdk::Option* kernelLoops = new streamsdk::Option;

    kernelLoops->_sVersion = "k";
    kernelLoops->_lVersion = "kernelLoops";
    kernelLoops->_description = "Number of loops in kernel";
    kernelLoops->_type = streamsdk::CA_ARG_INT;
    kernelLoops->_value = &nKLoops;

    sampleArgs->AddOption(kernelLoops);
    delete kernelLoops;

    streamsdk::Option* size = new streamsdk::Option;

    size->_sVersion = "x";
    size->_lVersion = "size";
    size->_description = "Size in bytes";
    size->_type = streamsdk::CA_ARG_INT;
    size->_value = &nBytes;

    sampleArgs->AddOption(size);
    delete size;

    streamsdk::Option* wavefronts = new streamsdk::Option;

    wavefronts->_sVersion = "w";
    wavefronts->_lVersion = "wavefronts";
    wavefronts->_description = "Number of wavefronts per compute unit";
    wavefronts->_type = streamsdk::CA_ARG_INT;
    wavefronts->_value = &numWavefronts;

    sampleArgs->AddOption(wavefronts);
    delete wavefronts;


    streamsdk::Option* inputFlags = new streamsdk::Option;

    inputFlags->_sVersion = "I";
    inputFlags->_lVersion = "inMemFlag";
    inputFlags->_description = "Memory flags for input buffer "
                         "\n\t\t 0 for CL_MEM_READ_ONLY"
                         "\n\t\t 1 CL_MEM_WRITE_ONLY"
                         "\n\t\t 2 CL_MEM_READ_WRITE"
                         "\n\t\t 3 CL_MEM_ALLOC_HOST_PTR"
                         "\n\t\t 4 CL_MEM_USE_PERSISTENT_MEM_AMD\n";
    inputFlags->_type = streamsdk::CA_ARG_INT;
    inputFlags->_value = &inFlagsValue;

    sampleArgs->AddOption(inputFlags);
    delete inputFlags;

    streamsdk::Option* noOverlapStr = new streamsdk::Option;

    noOverlapStr->_sVersion = "n";
    noOverlapStr->_lVersion = "noOverlap";
    noOverlapStr->_description = "Do not overlap memset() with kernel";
    noOverlapStr->_type = streamsdk::CA_NO_ARGUMENT;
    noOverlapStr->_value = &noOverlap;

    sampleArgs->AddOption(noOverlapStr);
    delete noOverlapStr;


    streamsdk::Option* pringLogStr = new streamsdk::Option;

    pringLogStr->_sVersion = "l";
    pringLogStr->_lVersion = "log";
    pringLogStr->_description = "Prints complete timing log";
    pringLogStr->_type = streamsdk::CA_NO_ARGUMENT;
    pringLogStr->_value = &printLog;

    sampleArgs->AddOption(pringLogStr);
    delete pringLogStr;

    return SDK_SUCCESS;
}

int 
TransferOverlap::parseExtraCommandLineOptions(int argc, char**argv)
{
    if(!parseCommandLine(argc, argv))
        return SDK_FAILURE;
    
    //Handle extra options which are not default in SDKUtils
    while(--argc) 
    {
        if(strcmp(argv[argc], "-I") == 0 ||  strcmp(argv[argc], "-inMemFlag") == 0)
        {
            int f = atoi( argv[ argc+1 ] );
            if( f < nFlags )
                inFlags |= memFlags[ f ].f;
        }
    }

    cl_mem_flags f = CL_MEM_ALLOC_HOST_PTR |
                     CL_MEM_USE_PERSISTENT_MEM_AMD |
                     CL_MEM_READ_ONLY |
                     CL_MEM_WRITE_ONLY |
                     CL_MEM_READ_WRITE;

    if( (inFlags & f) == 0 )
             inFlags |= CL_MEM_READ_ONLY | CL_MEM_USE_PERSISTENT_MEM_AMD;

    f  = CL_MEM_READ_ONLY |
         CL_MEM_WRITE_ONLY |
         CL_MEM_READ_WRITE;

    if( (inFlags & f) == 0 )
             inFlags |= CL_MEM_READ_ONLY;

    nSkip = nLoops > nSkip ? nSkip : 0;

    return SDK_SUCCESS;
}

int 
TransferOverlap::setup()
{
    int status = setupCL();
    if(status != SDK_SUCCESS)
        return status;

    return SDK_SUCCESS;
}

int 
TransferOverlap::verifyResultBuffer(cl_mem resultBuffer, bool firstLoop)
{
    cl_int   status;
    cl_event event;
    void *ptrResult = NULL;

    t.Reset(); 
    t.Start();

    ptrResult = (void*)clEnqueueMapBuffer(
                    queue,
                    resultBuffer,
                    CL_TRUE,
                    CL_MAP_READ,
                    0,
                    nBytesResult,
                    0, 
                    NULL,
                    NULL,
                    &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clEnqueueMapBuffer() failed.(resultBuffer)"))
        return SDK_FAILURE;

    t.Stop();
    timeLog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clEnqueueMapBuffer(MAP_WRITE)", 
        t.GetElapsedTime(), 
        nBytesResult, 
        1); 

    t.Reset(); 
    t.Start();

    cl_uint sum = 0;

    for( int i = 0; i < nThreads / MAX_WAVEFRONT_SIZE; i++ )
        sum += ((cl_uint*)ptrResult)[i];

    bool results;

    if( (sum != nBytes / sizeof(cl_uint)) && !firstLoop )
        results = false;
    else
        results = true;

    t.Stop();

    timeLog->Timer(
        "%32s  %lf s\n", "CPU reduction", 
        t.GetElapsedTime(), 
        nBytes, 
        1);
    
    if(results)
        timeLog->Msg( "%32s\n", "verification ok" );
    else
    {
        correctness = false;
        timeLog->Error( "%32s\n", "verification FAILED" );
        printf("Failed\n");
        return SDK_FAILURE;
    }
    
    t.Reset();
    t.Start();

    status = clEnqueueUnmapMemObject(
            queue,
            resultBuffer,
            (void *) ptrResult,
            0,
            NULL, 
            &event);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clEnqueueUnmapMemObject() failed.(resultBuffer)"))
        return SDK_FAILURE;
 
    status = clWaitForEvents(1, &event);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clWaitForEvents() failed."))
        return SDK_FAILURE;

    t.Stop();
    timeLog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clEnqueueUnmapMemObject()", 
        t.GetElapsedTime(), 
        nBytesResult,
        1);

    // Release event
    status = clReleaseEvent(event);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clReleaseEvent() failed.(event)"))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int 
TransferOverlap::launchKernel(cl_mem inputBuffer, cl_mem resultBuffer, unsigned char v)
{
    cl_uint vKernel = 0;
    cl_int status = 0;

    for(int i = 0; i < sizeof(cl_uint); i++)
       vKernel |= v << (i * 8);

    nItemsPerThread = nItems / nThreads;

    globalWorkSize = nThreads;
    localWorkSize = MAX_WAVEFRONT_SIZE;

    status = clSetKernelArg(
                readKernel, 
                0, 
                sizeof(void *),
                (void *)&inputBuffer);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clSetKernelArg() failed.(inputBuffer)"))
        return SDK_FAILURE;

    status = clSetKernelArg(
                readKernel, 
                1, 
                sizeof(void *),
                (void *) &resultBuffer);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clSetKernelArg() failed.(resultBuffer)"))
        return SDK_FAILURE;

    status = clSetKernelArg( readKernel, 
                2, 
                sizeof(cl_uint), 
                (void *)&nItemsPerThread);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clSetKernelArg() failed.(nItemsPerThread)"))
        return SDK_FAILURE;

    status = clSetKernelArg(
                readKernel,
                3,
                sizeof(cl_uint),
                (void *) &vKernel);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clSetKernelArg() failed.(vKernel)"))
        return SDK_FAILURE;

    status = clSetKernelArg(
                readKernel, 
                4, 
                sizeof(cl_uint), 
                (void *) &nKLoops);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clSetKernelArg() failed.(nKLoops)"))
        return SDK_FAILURE;

    cl_event  event;
    cl_event *evPtr;

    if( noOverlap )
        evPtr = &event;
    else
        evPtr = NULL;

    t.Reset(); 
    t.Start();

    status = clEnqueueNDRangeKernel(
            queue,
            readKernel,
            1,
            NULL,
            &globalWorkSize,
            &localWorkSize,
            0, 
            NULL, 
            evPtr);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clEnqueueNDRangeKernel() failed."))
        return SDK_FAILURE;

    if(noOverlap)
    {
        status = clWaitForEvents(1, evPtr);
        if(!sampleCommon->checkVal(
                status,
                CL_SUCCESS, 
                "clWaitForEvents() failed."))
            return SDK_FAILURE;
    }
    else
    {
        status = clFlush(queue);
        if(!sampleCommon->checkVal(
                status,
                CL_SUCCESS, 
                "clFlush() failed."))
            return SDK_FAILURE;
    }

    t.Stop();
    timeLog->Timer(
        "%32s  %lf s\n", "clEnqueueNDRangeKernel()", 
        t.GetElapsedTime(), 
        nBytes, 
        nKLoops);

    // Release the event
    if(noOverlap)
    {
        status = clReleaseEvent(*evPtr);
        if(!sampleCommon->checkVal(
                status,
                CL_SUCCESS, 
                "clReleaseEvent() failed.(*evPtr)"))
            return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}

void*
TransferOverlap::launchMapBuffer(cl_mem buffer, cl_event *mapEvent)
{
    cl_int status;
    void *ptr = NULL;

    t.Reset(); 
    t.Start();

    ptr = (void *)clEnqueueMapBuffer(
                    queue,
                    buffer,
                    CL_FALSE,
                    CL_MAP_WRITE,
                    0,
                    nBytes,
                    0,
                    NULL,
                    mapEvent,
                    &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clEnqueueMapBuffer() failed.(buffer)"))
        return NULL;

    status = clFlush(queue);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clFlush() failed."))
        return NULL;

    t.Stop();
    timeLog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clEnqueueMapBuffer(MAP_WRITE)", 
        t.GetElapsedTime(), 
        nBytes,
        1);

    return ptr;
}

int
TransferOverlap::fillBuffer(cl_mem buffer, cl_event *mapEvent, void *ptr, unsigned char v)
{
    cl_int status;
    cl_event event;

    t.Reset();
    t.Start();

    status = clWaitForEvents(1, mapEvent);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clWaitForEvents() failed."))
        return SDK_FAILURE;

    memset(ptr, v, nBytes);

    t.Stop();
    timeLog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clWaitForEvents() + memset()", 
        t.GetElapsedTime(), 
        nBytes, 
        1);    

    // Release event
    status = clReleaseEvent(*mapEvent);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clReleaseEvent() failed.(mapEvent)"))
        return SDK_FAILURE;

    t.Reset();
    t.Start();

    status = clEnqueueUnmapMemObject(
                queue,
                buffer,
                (void *) ptr,
                0,
                NULL,
                &event);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clEnqueueUnmapMemObject() failed.(buffer)"))
        return SDK_FAILURE;
 
    status = clWaitForEvents(1, &event);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clWaitForEvents() failed."))
        return SDK_FAILURE;

    t.Stop();
    timeLog->Timer(
        "%32s  %lf s %8.2lf GB/s\n", "clEnqueueUnmapMemObject()",
        t.GetElapsedTime(),
        nBytes,
        1);

	// Release event
    status = clReleaseEvent(event);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clReleaseEvent() failed.(event)"))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int 
TransferOverlap::runOverlapTest()
{
    int nl = nLoops;
    int status = SDK_SUCCESS;

    void *inPtr1 = NULL;
    void *inPtr2 = NULL;

    cl_event lastBuf1MapEvent;
    cl_event lastBuf2MapEvent;

    bool firstLoop = true;

    // Start with inputBuffer1 mapped
    inPtr1 = (void *)clEnqueueMapBuffer(
                            queue,
                            inputBuffer1,
                            CL_FALSE,
                            CL_MAP_WRITE,
                            0,
                            nBytes,
                            0,
                            NULL,
                            &lastBuf1MapEvent,
                            &status);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clEnqueueMapBuffer() failed.(inputBuffer1)"))
        return SDK_FAILURE;

    status = clFlush(queue);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clFlush() failed."))
        return SDK_FAILURE;

    CPerfCounter lt;

    lt.Reset(); 
    lt.Start();

    while(nl--)
    {
       timeLog->loopMarker();

       unsigned char v = (unsigned char)(nl & 0xff);

       // 1. Host acquires and fills inputBuffer1. Unless this
       // is the first loop, this happens concurrently with the
       // preceding kernel execution.

       timeLog->Msg( "\n%s\n\n", "Acquire and fill: inputBuffer1" );

       status = fillBuffer(inputBuffer1, &lastBuf1MapEvent, inPtr1, v);
       if(status != SDK_SUCCESS)
           return SDK_FAILURE;
    
       // This is a CPU/GPU synchronization point, as all commands in the
       // in-order queue before the preceding cl*Unmap() are now finished.
       // We can accurately sample the per-loop timer here.

       lt.Stop();
       timeLog->Timer(
           "\n%s %f s\n", "Loop time", 
           lt.GetElapsedTime(), 
           nBytes,
           1);
       lt.Reset();
       lt.Start();

       // 2. Launch map of inputBuffer2. The map needs to precede
       // the next kernel launch in the in-order queue, otherwise waiting
       // for the map to finish would also wait for the kernel to
       // finish.

       timeLog->Msg("\n%s\n\n", "Launch map: inputBuffer2");

       inPtr2 = launchMapBuffer(inputBuffer2, &lastBuf2MapEvent);
       if(inPtr2 == NULL)
           return SDK_FAILURE;

       // Verify result of kernel for inputBuffer2

       timeLog->Msg("\n%s\n\n", "Verify: resultBuffer2");

       status = verifyResultBuffer(resultBuffer2, firstLoop);
       if(status != SDK_SUCCESS)
           return SDK_FAILURE;

       // 3. Asynchronous launch of kernel for inputBuffer1

       timeLog->Msg("\n%s\n\n", "Launch GPU kernel: inputBuffer1");

       status = launchKernel(inputBuffer1, resultBuffer1, v);
       if(status != SDK_SUCCESS)
           return SDK_FAILURE;

       // 4. Host acquires and fills inputBuffer2. This happens
       // concurrently with the preceding kernel execution.

       timeLog->Msg("\n%s\n\n", "Acquire and fill: inputBuffer2");

       status = fillBuffer(inputBuffer2, &lastBuf2MapEvent, inPtr2, v); 
       if(status != SDK_SUCCESS)
           return SDK_FAILURE;

       // 5. Launch map of inputBuffer1. The map needs to precede
       // the next kernel launch in the in-order queue, otherwise waiting
       // for the map to finish would also wait for the kernel to
       // finish.

       timeLog->Msg("\n%s\n\n", "Launch map: inputBuffer1");

       inPtr1 = launchMapBuffer(inputBuffer1, &lastBuf1MapEvent);
       if(inPtr1 == NULL)
           return SDK_FAILURE;

       // Verify result of kernel for inputBuffer1

       timeLog->Msg("\n%s\n\n", "Verify: resultBuffer1");

       status = verifyResultBuffer(resultBuffer1, firstLoop);
       if(status != SDK_SUCCESS)
           return SDK_FAILURE;

       // 6. Asynchronous launch of kernel for inputBuffer2

       timeLog->Msg("\n%s\n\n", "Launch GPU kernel: inputBuffer2");

       status = launchKernel(inputBuffer2, resultBuffer2, v);
       if(status != SDK_SUCCESS)
           return SDK_FAILURE;

       timeLog->Msg("%s\n", "");

       firstLoop = false;
    }

    status = clFinish(queue);
    if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS, 
            "clFlush() failed."))
        return SDK_FAILURE;

    return SDK_SUCCESS;
}

int 
TransferOverlap::run()
{
    CPerfCounter gt;
    gt.Reset();
    gt.Start();

    int status = runOverlapTest();
    if(status != SDK_SUCCESS)
        return status;

    gt.Stop();
    testTime = gt.GetElapsedTime();

    if(verify)
    {
        if(correctness)
        {
            printf("Passed!!\n");
            return SDK_SUCCESS;
        }
        else
        {
            printf("Failed\n");
            return SDK_FAILURE;
        }
    }

    return SDK_SUCCESS;
}

void 
TransferOverlap::printStats()
{

    if(!quiet && !verify)
    {
#ifdef _WIN32
        printf( "Build:               _WINxx" ); 
#ifdef _DEBUG
        printf( " DEBUG");
#else
        printf( " release");
#endif
        printf( "\n" );
#else
#ifdef NDEBUG
        printf("Build:               release\n");
#else
        printf("Build:               DEBUG\n");
#endif
#endif

        printf("GPU work items:        %d\n", nThreads);
        printf("Buffer size:           %d\n", nBytes);
        printf("Timing loops:          %d\n", nLoops);
        printf("Kernel loops:          %d\n", nKLoops);
        printf("Wavefronts/SIMD:       %d\n", numWavefronts);
        printf("memset/kernel overlap: %s\n", noOverlap ? "no" : "yes");
        printf("inputBuffer:           " );

        for( int i = 0; i < sizeof(memFlags) / sizeof(memFlags[0]); i++)
           if(inFlags & memFlags[i].f)
              printf("%s ", memFlags[i].s);

        printf("\n");

        if(printLog) 
        {
           printf("\nLOOP ITERATIONS\n");
           printf("---------------\n\n");

           timeLog->printLog();
        }

        printf(
            "\nAVERAGES (over loops %d - %d, use -l to show complete log)\n",
            nSkip, 
            nLoops-1);
        printf("--------\n\n");

        timeLog->printSummary(nSkip);

        printf("\nComplete test time: %f s\n\n", testTime);
    }
}

int 
TransferOverlap::verifyResults()
{
    return SDK_SUCCESS;
}

int 
TransferOverlap::cleanup()
{
    /* Releases OpenCL resources (Context, Memory etc.) */
    cl_int status;

    if(inputBuffer1)
    {
        status = clReleaseMemObject(inputBuffer1);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(inputBuffer2)
    {
        status = clReleaseMemObject(inputBuffer2);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(resultBuffer1)
    {
        status = clReleaseMemObject(resultBuffer1);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    if(resultBuffer2)
    {
        status = clReleaseMemObject(resultBuffer2);
        if(!sampleCommon->checkVal(
            status,
            CL_SUCCESS,
            "clReleaseMemObject failed."))
            return SDK_FAILURE;
    }

    status = clReleaseKernel(readKernel);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseProgram failed."))
        return SDK_FAILURE;

    status = clReleaseKernel(writeKernel);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseProgram failed."))
        return SDK_FAILURE;

    status = clReleaseProgram(program);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseProgram failed."))
        return SDK_FAILURE;
 
    status = clReleaseCommandQueue(queue);
     if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseCommandQueue failed."))
        return SDK_FAILURE;

    status = clReleaseContext(context);
    if(!sampleCommon->checkVal(
        status,
        CL_SUCCESS,
        "clReleaseContext failed."))
        return SDK_FAILURE;


    delete timeLog;
    /* release device list */
    if(devices)
        free(devices);

    return SDK_SUCCESS;
}

int 
main(int argc, char * argv[])
{
    TransferOverlap clTransferOverlap("OpenCL Constant Bandwidth");

    clTransferOverlap.initialize();
    if(clTransferOverlap.parseExtraCommandLineOptions(argc, argv))
        return SDK_FAILURE;


    if(clTransferOverlap.isDumpBinaryEnabled())
    {
        return clTransferOverlap.genBinaryImage();
    }
    else
    {
        int state = clTransferOverlap.setup();
        if(state != SDK_SUCCESS)
        {
            if(state == SDK_EXPECTED_FAILURE)
                return SDK_SUCCESS;
            else
                return SDK_FAILURE;
        }

        state = clTransferOverlap.run();
        if(state == SDK_FAILURE)
          return SDK_FAILURE;

        clTransferOverlap.printStats();
        if(clTransferOverlap.cleanup()==SDK_FAILURE)
          return SDK_FAILURE;
    }

    return SDK_SUCCESS;
}
